
# Copyright 2017 Jaap Karssenberg <jaap.karssenberg@gmail.com>

# To install this plugin:
# 1. in zim go to the preferences dialog, plugins tab
# 2. click "open plugin folder"
#    this should open ~/.local/share/zim/plugins or similar
# 3. copy this file to the folder presented
# 4. restart zim (quit and then start again, don't just close the window)
# 5. go back to the plugins tab, plugin should now show up

# TODO: at this moment the plugin does not restore the menubar if disabled,
#       requires a restart of zim. If the plugin fails but the menubar is gone,
#       need to manually edit "preferences.conf" to remove the plugin; then
#       restart zim.

# TODO: test initialization of global items (about, prefs, ..)

# TODO: hide "help" menu from the menubar ?

from gi.repository import Gtk

from zim.gui.mainwindow import MainWindowExtension

from zim.plugins import PluginClass

try:
	import gi
	gi.require_version('GtkosxApplication', '1.0')
	from gi.repository import GtkosxApplication
except (ValueError, ImportError) as error:
	GtkosxApplication = None


if False: #pragma: no cover

	# This code defines a "mock" object for testing this plugin on systems
	# that do not actually have "GtkosxApplication", should be disabled for
	# production code.
	# Switch by setting above statement to "True" or "False"

	class MockOSXAppModule(object):

		@staticmethod
		def Application():
			return MockOSXAppObject()

	class MockOSXAppObject(object):

		def __getattr__(self, name):
			def method(*a):
				print(">>> OSX call:", name, a)
			return method

	if GtkosxApplication is None:
		GtkosxApplication = MockOSXAppModule


if GtkosxApplication:
	# Global for all notebooks / windows, once per process
	_global_osx_application = GtkosxApplication.Application()
	_global_items_initialized = False
else:
	_global_osx_application = None
	_global_items_initialized = False



class OSXmenubarPlugin(PluginClass):
	# This object just provides some information for the plugin manager
	# no real logic happening here.

	plugin_info = {
		'name': _('macOS Menubar'), # T: plugin name
		'description': _('This plugin provides a macOS menubar for zim.'), # T: plugin description
		'author': 'Brecht Machiels, Jaap Karssenberg',
		'help': 'Plugins:macOS Menubar'
	}

	@classmethod
	def check_dependencies(klass):
		# "is_ok" must be True, else won't be able to select the plugin in
		# the plugin manager
		is_ok = GtkosxApplication is not None
		return is_ok, [('GtkosxApplication', is_ok, True)]



class OSXMenuBarMainWindowExtension(MainWindowExtension):
	# This object is created once for each "main window", this means once for
	# each notebook opened in zim. If this is the first window, also do
	# global intialization, else just capture the menubar and keep it ourselves.
	# We hook to the signal that a window has received focus and on that signal
	# insert the menubar for that window. So may change often when switching
	# windows.

	def __init__(self, plugin, window):
		MainWindowExtension.__init__(self, plugin, window)

		# Define OS X menu bar for this window and remove menubar from window
		self.menubar = self.window.menubar
		self.window._zim_window_main.remove(self.menubar) # XXX - use private arg, should patch Window.remove() instead ...

		# Define global items - one time action for process
		global _global_osx_application
		_global_osx_application.set_use_quartz_accelerators(False)

		quit = self.window.uimanager.get_widget('/menubar/file_menu/quit')
		quit.hide()

		_global_osx_application.set_menu_bar(self.menubar)
		help = self.window.uimanager.get_widget('/menubar/help_menu')
		_global_osx_application.set_help_menu(help)

		about = self.window.uimanager.get_widget('/menubar/help_menu/show_about')
		_global_osx_application.insert_app_menu_item(about, 0)

		prefs = self.window.uimanager.get_widget('/menubar/edit_menu/show_preferences')
		_global_osx_application.insert_app_menu_item(Gtk.SeparatorMenuItem(), 1)
		_global_osx_application.insert_app_menu_item(prefs, 2)
		_global_osx_application.insert_app_menu_item(Gtk.SeparatorMenuItem(), 3)

		_global_osx_application.ready()



# Copyright 2008-2023 Jaap Karssenberg <jaap.karssenberg@gmail.com>

import logging

from gi.repository import Gtk

import os
import zim.datetimetz as datetime

from zim.errors import Error
from zim.config import ConfigManager
from zim.formats import get_format
from zim.notebook import HRef, PageNotFoundError
from zim.parse.links import link_type
from zim.templates.expression import ExpressionFunction
from zim.gui.widgets import Dialog, FileDialog, ErrorDialog, BrowserTreeView, ScrolledWindow
from zim.gui.base.images import image_file_get_dimensions
from zim.gui.applications import edit_config_file, open_folder


logger = logging.getLogger('zim.gui.pageview')



class InsertDateDialog(Dialog):
	'''Dialog to insert a date-time in the page'''

	FORMAT_COL = 0 # format string
	DATE_COL = 1 # strfime rendering of the format

	def __init__(self, parent, buffer, notebook, page):
		Dialog.__init__(
			self,
			parent,
			_('Insert Date and Time'), # T: Dialog title
			button=_('_Insert'), # T: Button label
			use_default_button=True
		)
		self.buffer = buffer
		self.notebook = notebook
		self.page = page
		self.date = datetime.now()

		self.uistate.setdefault('lastusedformat', '')
		self.uistate.setdefault('linkdate', False)

		## Add Calendar widget
		from zim.plugins.journal import Calendar # FIXME put this in zim.gui.widgets

		label = Gtk.Label()
		label.set_markup('<b>' + _("Date") + '</b>') # T: label in "insert date" dialog
		label.set_alignment(0.0, 0.5)
		self.vbox.pack_start(label, False, False, 0)

		self.calendar = Calendar()
		self.calendar.set_display_options(
			Gtk.CalendarDisplayOptions.SHOW_HEADING |
			Gtk.CalendarDisplayOptions.SHOW_DAY_NAMES |
			Gtk.CalendarDisplayOptions.SHOW_WEEK_NUMBERS)
		self.calendar.connect('day-selected', lambda c: self.set_date(c.get_date()))
		self.vbox.pack_start(self.calendar, False, True, 0)

		## Add format list box
		label = Gtk.Label()
		label.set_markup('<b>' + _("Format") + '</b>') # T: label in "insert date" dialog
		label.set_alignment(0.0, 0.5)
		self.vbox.pack_start(label, False, False, 0)

		model = Gtk.ListStore(str, str) # FORMAT_COL, DATE_COL
		self.view = BrowserTreeView(model)
		self.vbox.pack_start(ScrolledWindow(self.view), True, True, 0)

		cell_renderer = Gtk.CellRendererText()
		column = Gtk.TreeViewColumn('_date_', cell_renderer, text=1)
		self.view.append_column(column)
		self.view.set_headers_visible(False)
		self.view.connect('row-activated',
			lambda *a: self.response(Gtk.ResponseType.OK))

		## Add Link checkbox and Edit button
		self.linkbutton = Gtk.CheckButton.new_with_mnemonic(_('_Link to date'))
			# T: check box in InsertDate dialog
		self.linkbutton.set_active(self.uistate['linkdate'])
		self.vbox.pack_start(self.linkbutton, False, True, 0)

		button = Gtk.Button.new_with_mnemonic(_('_Edit')) # T: Button label
		button.connect('clicked', self.on_edit)
		self.action_area.add(button)
		self.action_area.reorder_child(button, 1)

		## Setup data
		self.load_file()
		self.set_date(self.date)

	def load_file(self):
		lastused = None
		model = self.view.get_model()
		model.clear()
		file = ConfigManager.get_config_file('dates.list')
		for line in file.readlines():
			line = line.strip()
			if not line or line.startswith('#'):
				continue
			try:
				format = line
				iter = model.append((format, format))
				if format == self.uistate['lastusedformat']:
					lastused = iter
			except:
				logger.exception('Could not parse date: %s', line)

		if len(model) == 0:
			# file not found ?
			model.append(("%c", "%c"))

		if not lastused is None:
			path = model.get_path(lastused)
			self.view.get_selection().select_path(path)

	def set_date(self, date):
		self.date = date

		def update_date(model, path, iter):
			format = model[iter][self.FORMAT_COL]
			try:
				string = datetime.strftime(format, date)
			except ValueError:
				string = 'INVALID: ' + format
			model[iter][self.DATE_COL] = string

		model = self.view.get_model()
		model.foreach(update_date)

		link = date.strftime('%Y-%m-%d') # YYYY-MM-DD
		self.link = self.notebook.suggest_link(self.page, link)
		self.linkbutton.set_sensitive(not self.link is None)

	#def run(self):
		#self.view.grab_focus()
		#Dialog.run(self)

	def save_uistate(self):
		model, iter = self.view.get_selection().get_selected()
		if iter:
			format = model[iter][self.FORMAT_COL]
			self.uistate['lastusedformat'] = format
		self.uistate['linkdate'] = self.linkbutton.get_active()

	def on_edit(self, button):
		file = ConfigManager.get_config_file('dates.list') # XXX
		if edit_config_file(self, file):
			self.load_file()

	def do_response_ok(self):
		model, iter = self.view.get_selection().get_selected()
		if iter:
			text = model[iter][self.DATE_COL]
		else:
			text = model[0][self.DATE_COL]

		if self.link and self.linkbutton.get_active():
			self.buffer.insert_link_at_cursor(text, self.link.name)
		else:
			self.buffer.insert_at_cursor(text)

		return True


class InsertImageDialog(FileDialog):
	'''Dialog to insert an image in the page'''

	def __init__(self, parent, buffer, notebook, path, file=None):
		FileDialog.__init__(
			self, parent, _('Insert Image'), Gtk.FileChooserAction.OPEN)
			# T: Dialog title

		self.buffer = buffer
		self.notebook = notebook
		self.path = path

		self.uistate.setdefault('attach_inserted_images', False)
		self.uistate.setdefault('last_image_folder', None, check=str)

		self.add_shortcut(notebook, path)
		self.add_filter_images()

		checkbox = Gtk.CheckButton.new_with_mnemonic(_('Attach image first'))
			# T: checkbox in the "Insert Image" dialog
		checkbox.set_active(self.uistate['attach_inserted_images'])
		self.filechooser.set_extra_widget(checkbox)

		if file:
			self.set_file(file)
		else:
			self.load_last_folder()

	def do_response_ok(self):
		file = self.get_file()
		if file is None:
			return False

		try:
			image_file_get_dimensions(file.path)
		except AssertionError:
			ErrorDialog(self, _('File type not supported: %s') % file.mimetype()).run()
				# T: Error message when trying to insert a not supported file as image
			return False

		self.save_last_folder()

		# Similar code in AttachFileDialog
		checkbox = self.filechooser.get_extra_widget()
		self.uistate['attach_inserted_images'] = checkbox.get_active()
		if self.uistate['attach_inserted_images']:
			folder = self.notebook.get_attachments_dir(self.path)
			if not file.ischild(folder):
				file = attach_file(self, self.notebook, self.path, file)
				if file is None:
					return False # Cancelled overwrite dialog

		src = self.notebook.relative_filepath(file, self.path) or file.uri
		self.buffer.insert_image_at_cursor(file, src)
		return True


class AttachFileDialog(FileDialog):

	def __init__(self, parent, buffer, notebook, path, file=None):
		assert path, 'Need a page here'
		FileDialog.__init__(self, parent, _('Attach File'), multiple=True) # T: Dialog title
		self.buffer = buffer
		self.notebook = notebook
		self.path = path

		dir = notebook.get_attachments_dir(path)
		if dir is None:
			ErrorDialog(self, _('Page "%s" does not have a folder for attachments') % self.path)
				# T: Error dialog - %s is the full page name
			raise Exception('Page "%s" does not have a folder for attachments' % self.path)

		self.add_shortcut(notebook, path)
		if file:
			self.set_file(file)
		else:
			self.load_last_folder()

	def do_response_ok(self):
		files = self.get_files()
		if not files:
			return False

		self.save_last_folder()

		inserted = False
		last = len(files) - 1
		for i, file in enumerate(files):
			file = attach_file(self, self.notebook, self.path, file)
			if file is not None:
				inserted = True
				text = self.notebook.relative_filepath(file, path=self.path)
				self.buffer.insert_link_at_cursor(text, href=text)
				if i != last:
					self.buffer.insert_at_cursor(' ')

		return inserted # If nothing is inserted, return False and do not close dialog


def attach_file(widget, notebook, path, file, force_overwrite=False):
	folder = notebook.get_attachments_dir(path)
	if folder is None:
		raise Error('%s does not have an attachments dir' % path)

	dest = folder.file(file.basename)
	if dest.exists() and not force_overwrite:
		dialog = PromptExistingFileDialog(widget, dest)
		dest = dialog.run()
		if dest is None:
			return None	# dialog was cancelled
		elif dest.exists():
			dest.remove()

	file.copyto(dest)
	return dest


class PromptExistingFileDialog(Dialog):
	'''Dialog that is used e.g. when a file should be attached to zim,
	but a file with the same name already exists in the attachment
	directory. This Dialog allows to suggest a new name or overwrite
	the existing one.

	For this dialog C{run()} will return either the original file
	(for overwrite), a new file, or None when the dialog was canceled.
	'''

	def __init__(self, widget, file):
		Dialog.__init__(self, widget, _('File Exists'), buttons=None) # T: Dialog title
		self.add_help_text( _('''\
A file with the name <b>"%s"</b> already exists.
You can use another name or overwrite the existing file.''' % file.basename),
		) # T: Dialog text in 'new filename' dialog
		self.folder = file.parent()
		self.old_file = file

		suggested_filename = self.folder.new_file(file.basename).basename
		self.add_form((
				('name', 'string', _('Filename')), # T: Input label
			), {
				'name': suggested_filename
			}
		)
		self.form.widgets['name'].set_check_func(self._check_valid)

		# all buttons are defined in this class, to get the ordering right
		# [show folder]      [overwrite] [cancel] [ok]
		button = Gtk.Button.new_with_mnemonic(_('_Browse')) # T: Button label
		button.connect('clicked', self.do_show_folder)
		self.action_area.add(button)
		self.action_area.set_child_secondary(button, True)

		button = Gtk.Button.new_with_mnemonic(_('Overwrite')) # T: Button label
		button.connect('clicked', self.do_response_overwrite)
		self.add_action_widget(button, Gtk.ResponseType.NONE)

		self.add_button(_('_Cancel'), Gtk.ResponseType.CANCEL) # T: Button label
		self.add_button(_('_OK'), Gtk.ResponseType.OK) # T: Button label
		self._no_ok_action = False

		self.form.widgets['name'].connect('focus-in-event', self._on_focus)

	def _on_focus(self, widget, event):
		# filename length without suffix
		length = len(os.path.splitext(widget.get_text())[0])
		widget.select_region(0, length)

	def _check_valid(self, filename):
		# Only valid when same dir and does not yet exist
		file = self.folder.file(filename)
		return file.ischild(self.folder) and not file.exists()

	def do_show_folder(self, *a):
		open_folder(self, self.folder)

	def do_response_overwrite(self, *a):
		logger.info('Overwriting %s', self.old_file.path)
		self.result = self.old_file

	def do_response_ok(self):
		if not self.form.widgets['name'].get_input_valid():
			return False

		newfile = self.folder.file(self.form['name'])
		logger.info('Selected %s', newfile.path)
		assert newfile.ischild(self.folder) # just to be real sure
		assert not newfile.exists() # just to be real sure
		self.result = newfile
		return True


class EditImageDialog(Dialog):
	'''Dialog to edit properties of an embedded image'''

	def __init__(self, parent, buffer, notebook, path):
		Dialog.__init__(self, parent, _('Edit Image')) # T: Dialog title
		self.buffer = buffer
		self.notebook = notebook
		self.path = path

		iter = buffer.get_iter_at_mark(buffer.get_insert())
		image_data = self.buffer.get_image_data(iter)
		if image_data is None:
			iter.backward_char()
			image_data = self.buffer.get_image_data(iter)
			assert image_data, 'No image found'
		self._image_data = image_data.copy()
		self._iter = iter.get_offset()

		src = image_data['src']
		if '?' in src:
			i = src.find('?')
			src = src[:i]
		href = image_data.get('href', '')
		anchor = image_data.get('id', '')
		self.add_form([
				('file', 'image', _('Location')), # T: Input in 'edit image' dialog
				('href', 'link', _('Link to'), path), # T: Input in 'edit image' dialog
				('width', 'int', _('Width'), (0, 1)), # T: Input in 'edit image' dialog
				('height', 'int', _('Height'), (0, 1)), # T: Input in 'edit image' dialog
				('anchor', 'string', _('Id'))
			],
			{'file': src, 'href': href, 'anchor': anchor}
			# range for width and height are set in set_ranges()
		)
		self.form.widgets['file'].set_use_relative_paths(notebook, path)
		self.form.widgets['file'].allow_empty = False
		self.form.widgets['file'].show_empty_invalid = True
		self.form.widgets['file'].update_input_valid()

		reset_button = Gtk.Button.new_with_mnemonic(_('_Reset Size'))
			# T: Button in 'edit image' dialog
		hbox = Gtk.HBox()
		hbox.pack_end(reset_button, False, True, 0)
		self.vbox.add(hbox)

		reset_button.connect_object('clicked',
			self.__class__.reset_dimensions, self)
		self.form.widgets['file'].connect_object('changed',
			self.__class__.do_file_changed, self)
		self.form.widgets['width'].connect_object('value-changed',
			self.__class__.do_width_changed, self)
		self.form.widgets['height'].connect_object('value-changed',
			self.__class__.do_height_changed, self)

		# Init ranges based on original
		self.reset_dimensions()

		# Set current scale if any
		if 'width' in image_data:
			self.form.widgets['width'].set_value(int(image_data['width']))
		elif 'height' in image_data:
			self.form.widgets['height'].set_value(int(image_data['height']))

	def reset_dimensions(self):
		self._image_data.pop('width', None)
		self._image_data.pop('height', None)
		width = self.form.widgets['width']
		height = self.form.widgets['height']
		file = self.form['file']
		try:
			if file is None:
				raise AssertionError
			w, h = image_file_get_dimensions(file.path) # can raise
		except:
			logger.warning('Could not get size for image: %s', file.path)
			width.set_sensitive(False)
			height.set_sensitive(False)
		else:
			width.set_sensitive(True)
			height.set_sensitive(True)
			self._block = True
			width.set_range(0, 4 * w)
			width.set_value(w)
			height.set_range(0, 4 * w)
			height.set_value(h)
			self._block = False
			self._ratio = float(w) / h

	def do_file_changed(self):
		# Prevent images becoming one pixel wide
		file = self.form['file']
		if file is None:
			return
		try:
			if self._image_data['width'] == 1:
				self.reset_dimensions()
		except KeyError:
			# width hasn't been set
			pass

	def do_width_changed(self):
		if hasattr(self, '_block') and self._block:
			return
		self._image_data.pop('height', None)
		self._image_data['width'] = int(self.form['width'])
		h = int(float(self._image_data['width']) / self._ratio)
		self._block = True
		self.form['height'] = h
		self._block = False

	def do_height_changed(self):
		if hasattr(self, '_block') and self._block:
			return
		self._image_data.pop('width', None)
		self._image_data['height'] = int(self.form['height'])
		w = int(self._ratio * float(self._image_data['height']))
		self._block = True
		self.form['width'] = w
		self._block = False

	def do_response_ok(self):
		file = self.form['file']
		if file is None:
			return False

		attrib = self._image_data
		attrib['src'] = self.notebook.relative_filepath(file, self.path) or file.uri

		href = self.form['href']
		if href:
			type = link_type(href)
			if type == 'file':
				# Try making the path relative
				linkfile = self.form.widgets['href'].get_file()
				href = self.notebook.relative_filepath(linkfile, self.path) or linkfile.uri
			attrib['href'] = href
		else:
			attrib.pop('href', None)

		id = self.form['anchor']
		if id:
			attrib['id'] = id
		else:
			attrib.pop('id', None)

		iter = self.buffer.get_iter_at_offset(self._iter)
		bound = iter.copy()
		bound.forward_char()
		with self.buffer.user_action:
			self.buffer.delete(iter, bound)
			self.buffer.insert_image_at_cursor(file, **attrib)
		return True


class InsertTextFromFileDialog(FileDialog):
	'''Dialog to insert text from an external file into the page'''

	def __init__(self, parent, buffer, notebook, page):
		FileDialog.__init__(
			self, parent, _('Insert Text From File'), Gtk.FileChooserAction.OPEN)
			# T: Dialog title
		self.load_last_folder()
		self.add_shortcut(notebook, page)
		self.buffer = buffer

	def do_response_ok(self):
		file = self.get_file()
		if file is None:
			return False
		parser = get_format('plain').Parser()
		tree = parser.parse(file.readlines())
		self.buffer.insert_parsetree_at_cursor(tree)
		self.save_last_folder()
		return True


class InsertLinkDialog(Dialog):
	'''Dialog to insert a new link in the page or edit properties of
	an existing link
	'''

	def __init__(self, parent, pageview):
		self.pageview = pageview
		href, text = self._get_link_from_buffer()

		if href:
			title = _('Edit Link') # T: Dialog title
		else:
			title = _('Insert Link') # T: Dialog title

		Dialog.__init__(self, parent, title, button=_('_Link'), help='Help:Links')  # T: Dialog button

		self.uistate.setdefault('short_links', pageview.notebook.config['Notebook']['short_links'])
		self.add_form(
			[
				('href', 'link', _('Link to'), pageview.page), # T: Input in 'insert link' dialog
				('text', 'string', _('Text')), # T: Input in 'insert link' dialog
				('short_links', 'bool', _('Prefer short link names for pages')), # T: Input in 'insert link' dialog
			], {
				'href': href,
				'text': text,
				'short_links': self.uistate['short_links'],
			},
			notebook=pageview.notebook
		)

		# Hook text entry to copy text from link when apropriate
		self.form.widgets['href'].connect('changed', self.on_href_changed)
		self.form.widgets['text'].connect('changed', self.on_text_changed)
		self.form.widgets['short_links'].connect('toggled', self.on_short_link_pref_changed)
		self._text_for_link = self._link_to_text(href)
		self._copy_text = self._text_for_link == text and not self._selected_text

	def _get_link_from_buffer(self):
		# Get link and text from the text buffer
		href, text = '', ''

		buffer = self.pageview.textview.get_buffer()
		if buffer.get_has_selection():
			buffer.strip_selection()
			link = buffer.get_has_link_selection()
		else:
			link = buffer.select_link()
			if not link:
				buffer.select_word()

		if buffer.get_has_selection():
			start, end = buffer.get_selection_bounds()
			text = start.get_text(end)
			self._selection_bounds = (start.get_offset(), end.get_offset())
				# Interaction in the dialog causes buffer to loose selection
				# maybe due to clipboard focus !??
				# Anyway, need to remember bounds ourselves.
			if link:
				href = link['href']
				self._selected_text = False
			else:
				href = text
				self._selected_text = True
		else:
			self._selection_bounds = None
			self._selected_text = False

		return href, text

	def on_href_changed(self, o):
		# Check if we can also update text
		self._text_for_link = self._link_to_text(self.form['href'])
		if self._copy_text:
			self.form['text'] = self._text_for_link
			self._copy_text = True # just to be sure

	def on_text_changed(self, o):
		# Check if we should stop updating text
		self._copy_text = self.form['text'] == self._text_for_link

	def on_short_link_pref_changed(self, o):
		self.on_href_changed(None)

	def _link_to_text(self, link):
		if not link:
			return ''
		elif self.form['short_links'] and link_type(link) == 'page':
				# Similar to 'short_links' notebook property but using uistate instead
				try:
					href = HRef.new_from_wiki_link(link)
				except ValueError:
					return ''
				else:
					return href.short_name()
		else:
			return link

	def do_response_ok(self):
		self.uistate['short_links'] = self.form['short_links']

		href = self.form['href']
		if not href:
			self.form.widgets['href'].set_input_valid(False)
			return False

		type = link_type(href)
		if type == 'file':
			# Try making the path relative
			try:
				file = self.form.widgets['href'].get_file()
				page = self.pageview.page
				notebook = self.pageview.notebook
				href = notebook.relative_filepath(file, page) or file.uri
			except:
				pass # E.g. malformed path

		text = self.form['text'] or href

		buffer = self.pageview.textview.get_buffer()
		with buffer.user_action:
			if self._selection_bounds:
				start, end = list(map(
					buffer.get_iter_at_offset, self._selection_bounds))
				buffer.delete(start, end)
			buffer.insert_link_at_cursor(text, href)

		return True


class WordCountDialog(Dialog):
	'''Dialog showing line, word, and character counts'''

	def __init__(self, pageview):
		Dialog.__init__(self, pageview,
			_('Word Count'), buttons=Gtk.ButtonsType.CLOSE) # T: Dialog title
		self.set_resizable(False)

		def count(buffer, bounds):
			start, end = bounds
			lines = end.get_line() - start.get_line() + 1
			chars = end.get_offset() - start.get_offset()

			strings = start.get_text(end).strip().split()
			non_space_chars = sum(len(s) for s in strings)

			words = 0
			iter = start.copy()
			while iter.compare(end) < 0:
				if iter.forward_word_end():
					words += 1
				elif iter.compare(end) == 0:
					# When end is end of buffer forward_end_word returns False
					words += 1
					break
				else:
					break

			return lines, words, chars, non_space_chars

		buffer = pageview.textview.get_buffer()
		buffercount = count(buffer, buffer.get_bounds())
		insert = buffer.get_iter_at_mark(buffer.get_insert())
		start = buffer.get_iter_at_line(insert.get_line())
		end = start.copy()
		end.forward_line()
		paracount = count(buffer, (start, end))
		if buffer.get_has_selection():
			selectioncount = count(buffer, buffer.get_selection_bounds())
		else:
			selectioncount = (0, 0, 0, 0)

		table = Gtk.Table(3, 4)
		table.set_row_spacings(5)
		table.set_col_spacings(12)
		self.vbox.add(table)

		plabel = Gtk.Label(label=_('Page')) # T: label in word count dialog
		alabel = Gtk.Label(label=_('Paragraph')) # T: label in word count dialog
		slabel = Gtk.Label(label=_('Selection')) # T: label in word count dialog
		wlabel = Gtk.Label(label='<b>' + _('Words') + '</b>:') # T: label in word count dialog
		llabel = Gtk.Label(label='<b>' + _('Lines') + '</b>:') # T: label in word count dialog
		clabel = Gtk.Label(label='<b>' + _('Characters') + '</b>:') # T: label in word count dialog
		dlabel = Gtk.Label(label='<b>' + _('Characters excluding spaces') + '</b>:') # T: label in word count dialog

		for label in (wlabel, llabel, clabel, dlabel):
			label.set_use_markup(True)
			label.set_alignment(0.0, 0.5)

		# Heading
		table.attach(plabel, 1, 2, 0, 1)
		table.attach(alabel, 2, 3, 0, 1)
		table.attach(slabel, 3, 4, 0, 1)

		# Lines
		table.attach(llabel, 0, 1, 1, 2)
		table.attach(Gtk.Label(label=str(buffercount[0])), 1, 2, 1, 2)
		table.attach(Gtk.Label(label=str(paracount[0])), 2, 3, 1, 2)
		table.attach(Gtk.Label(label=str(selectioncount[0])), 3, 4, 1, 2)

		# Words
		table.attach(wlabel, 0, 1, 2, 3)
		table.attach(Gtk.Label(label=str(buffercount[1])), 1, 2, 2, 3)
		table.attach(Gtk.Label(label=str(paracount[1])), 2, 3, 2, 3)
		table.attach(Gtk.Label(label=str(selectioncount[1])), 3, 4, 2, 3)

		# Characters
		table.attach(clabel, 0, 1, 3, 4)
		table.attach(Gtk.Label(label=str(buffercount[2])), 1, 2, 3, 4)
		table.attach(Gtk.Label(label=str(paracount[2])), 2, 3, 3, 4)
		table.attach(Gtk.Label(label=str(selectioncount[2])), 3, 4, 3, 4)

		# Characters excluding spaces
		table.attach(dlabel, 0, 1, 4, 5)
		table.attach(Gtk.Label(label=str(buffercount[3])), 1, 2, 4, 5)
		table.attach(Gtk.Label(label=str(paracount[3])), 2, 3, 4, 5)
		table.attach(Gtk.Label(label=str(selectioncount[3])), 3, 4, 4, 5)


from zim.notebook import update_parsetree_and_copy_images

class MoveTextDialog(Dialog):
	'''This dialog allows moving a selected text to a new page
	The idea is to allow "refactoring" of pages more easily.
	'''

	def __init__(self, pageview, notebook, page, buffer, navigation):
		assert buffer.get_has_selection(), 'No Selection present'
		# save selection bounds (see #1963)
		self.bounds = buffer.get_selection_bounds()

		Dialog.__init__(
			self,
			pageview,
			_('Move Text to Other Page'), # T: Dialog title
			button=_('_Move')  # T: Button label
		)
		self.pageview = pageview
		self.notebook = notebook
		self.page = page
		self.buffer = buffer
		self.navigation = navigation

		self.uistate.setdefault('link', True)
		self.uistate.setdefault('open_page', False)
		self.uistate.setdefault('short_links', pageview.notebook.config['Notebook']['short_links'])
		self.add_form([
			('page', 'page', _('Move text to'), page), # T: Input in 'move text' dialog
			('link', 'bool', _('Leave link to new page')), # T: Input in 'move text' dialog
			('short_links', 'bool', _('Prefer short link names for pages')), # T: Input in 'move text' dialog
			('open_page', 'bool', _('Open new page')), # T: Input in 'move text' dialog

		], self.uistate)

		self.form.widgets['link'].connect('toggled', self.on_link_changed)
		self.on_link_changed()

	def on_link_changed(self, *a):
		# Set short link names toggle sensitivity according to link leaving toggle
		self.form.widgets['short_links'].set_sensitive(self.form['link'])

	def do_response_ok(self):
		newpage = self.form['page']
		if not newpage:
			return False

		try:
			newpage = self.notebook.get_page(newpage)
		except PageNotFoundError:
			return False

		# Copy text
		if not self.bounds:
			ErrorDialog(self, _('No text selected')).run() # T: error message in "move selected text" action
			return False

		if not newpage.exists():
			template = self.notebook.get_template(newpage,
				{'place_cursor': ExpressionFunction(lambda: '')} # FIXME, ideal would be to insert new content at cursor location
			)
			newpage.set_parsetree(template)

		parsetree = self.buffer.get_parsetree(self.bounds)
		newtree = update_parsetree_and_copy_images(parsetree, self.notebook, self.page, newpage)

		newpage.append_parsetree(newtree)
		self.notebook.store_page(newpage)

		# Delete text (after copy was successfull..)
		self.buffer.delete(*self.bounds)

		# Save link format preference
		self.uistate['short_links'] = self.form['short_links']

		# Insert Link
		self.uistate['link'] = self.form['link']
		if self.form['link']:
			href = self.form.widgets['page'].get_text()  # TODO add method to Path "get_link" which gives rel path formatted correctly
			try:
				text = HRef.new_from_wiki_link(href).short_name() if self.uistate['short_links'] else href
			except ValueError:
				text = href
			self.buffer.insert_link_at_cursor(text, href)

		# Show page
		self.uistate['open_page'] = self.form['open_page']
		if self.form['open_page']:
			self.navigation.open_page(newpage)

		return True


class NewFileDialog(Dialog):

	def __init__(self, parent, basename):
		Dialog.__init__(self, parent, _('New File')) # T: Dialog title
		self.add_form((
			('basename', 'string', _('Name')), # T: input for new file name
		), {
			'basename': basename
		})

	def show_all(self):
		Dialog.show_all(self)

		# Select only first part of name
		# TODO - make this a widget type in widgets.py
		text = self.form.widgets['basename'].get_text()
		if '.' in text:
			name, ext = text.split('.', 1)
			self.form.widgets['basename'].select_region(0, len(name))

	def do_response_ok(self):
		self.result = self.form['basename']
		return True



# Copyright 2024 Radu Potop <radu@wooptoo.com>

import logging

from zim.plugins import PluginClass
from zim.plugins.base.imagegenerator import (
	ImageGeneratorClass,
	BackwardImageGeneratorObjectType,
)

from zim.newfs import LocalFile, TmpFile
from zim.applications import Application, ApplicationError

logger = logging.getLogger('zim.plugins.mermaiddiagrameditor')


def get_cmd(fmt):
	return ('mmdc', '-q', '-e', fmt, '-b', 'transparent', '-i')


class InsertMermaidDiagramPlugin(PluginClass):
	plugin_info = {
		'name': _('Insert Mermaid Diagram'),  # T: plugin name
		'description': _('''\
This plugin provides a Mermaid diagram editor for zim.
It allows easy editing of Mermaid diagrams.
'''),  # T: plugin description
		'help': 'Plugins:Mermaid Diagram Editor',
		'author': 'Radu Potop',
	}

	## Disabled because generating SVGs is buggy.
	## Sequence diagrams work fine, but Flowcharts and Class diagrams fail to
	## render labels.
	##
	## Bug report: https://github.com/mermaid-js/mermaid-cli/issues/691

	# plugin_preferences = (
	#	  # key, type, label, default
	#	  (
	#		  'prefer_svg',
	#		  'bool',
	#		  _('Generate diagrams in SVG format'),	 # T: plugin preference
	#		  supports_image_format('svg'),
	#	  ),
	# )

	@classmethod
	def check_dependencies(klass):
		has_diagcmd = Application(get_cmd('png')).tryexec()
		return has_diagcmd, [('mmdc', has_diagcmd, True)]


class BackwardMermaidDiagramImageObjectType(BackwardImageGeneratorObjectType):
	name = 'image+mermaid'
	label = _('Mermaid Diagram')  # T: menu item
	syntax = None
	scriptname = 'mermaid-diagram.mmd'


class MermaidDiagramGenerator(ImageGeneratorClass):
	@property
	def _pref_format(self):
		return 'png'

	@property
	def imagefile_extension(self):
		return '.' + self._pref_format

	@property
	def diagcmd(self):
		return get_cmd(self._pref_format)

	def __init__(self, plugin, notebook, page):
		ImageGeneratorClass.__init__(self, plugin, notebook, page)
		self.mmdfile = TmpFile('mermaid-diagram.mmd')
		self.mmdfile.touch()

	def generate_image(self, text):
		# Write to tmp file
		self.mmdfile.write(text)
		self.imgfile = LocalFile(self.mmdfile.path[:-4] + self.imagefile_extension)
		logger.debug('Writing diagram to temp file: %s', self.imgfile)

		# Call mmdc
		try:
			diag = Application(self.diagcmd)
			diag.run((self.mmdfile, '-o', self.imgfile))
		except ApplicationError as e:
			logger.debug('Generating diagram failed with error: %s', e)
			return None, None
		else:
			return self.imgfile, None

	def cleanup(self):
		self.mmdfile.remove()
		try:
			self.imgfile.remove()
		except AttributeError:
			logger.debug('Closed dialog before generating image, nothing to remove')



# Copyright 2011 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module contains classes to work with simple tree structure
representation of formatted text.
'''

from .tokenlist import TEXT, END
from .builder import Builder



class SimpleTreeElement(list):
	'''This class represents (part of) a simple tree structure
	of formatted text.

	The attributes `tag` and `attrib` contain the element name
	and a dictionary with properties. If the element has no
	properties `attrib` can be `None` instead.

	It can be accessed as a list to find the child elements.
	All child elements are either a string with plain text
	or a nested `SimpleTreeElement`.

	The L{SimpleTreeBuilder} and L{SimpleTreeParser} classes can help
	to construct and process these tree structures.
	See also L{simpletree_to_tokens()} and L{tokens_to_simpletree()}
	'''

	__slots__ = ('tag', 'attrib')

	def __init__(self, tag, attrib=None, children=None):
		self.tag = tag
		self.attrib = attrib
		if children:
			self.extend(children)

	def set(self, key, value):
		'''Set attribute'''
		if self.attrib is None:
			self.attrib = {}
		self.attrib[key] = value

	def get(self, key, default=None):
		'''Get attribute'''
		if self.attrib:
			return self.attrib.get(key, default)
		else:
			return default

	def __eq__(self, other):
		if self.tag == other.tag \
		and self.attrib == other.attrib \
		and len(self) == len(other):
			return all(s == o for s, o in zip(self, other))
		else:
			return False

	def __repr__(self):
		if len(self) > 0:
			return '<%s:\n%s>' % (self.__class__.__name__, self.pprint(level=1))
		else:
			return '<%s: %s>' % (self.__class__.__name__, self.pprint(level=0).strip())

	def __str__(self):
		return self.__repr__()

	def pprint(self, level=0):
		'''Returns pretty-printed text representation'''
		prefix = '  ' * level
		if len(self) > 0:
			lines = [prefix + '%s %r [\n' % (self.tag, self.attrib)]
			for item in self:
				if isinstance(item, SimpleTreeElement):
					lines.append(item.pprint(level=level + 1))
				elif isinstance(item, str):
					for line in item.splitlines(True):
						lines.append(prefix + '  %r\n' % line)
				else:
					lines.append(prefix + '  %r\n' % item)
			lines.append(prefix + ']\n')
			return ''.join(lines)
		else:
			return prefix + '%s %r []\n' % (self.tag, self.attrib)


class SimpleTreeBuilder(Builder):
	'''Builder class that builds a tree of L{SimpleTreeElement}s'''

	def __init__(self, elementfactory=SimpleTreeElement):
		self.elementfactory = elementfactory
		self.toplevel = []
		self.stack = [self.toplevel]

	def get_root(self):
		'''Return top level element
		raises `AssertionError` if content not complete (e.g. unclosed tags) or no single toplevel element
		'''
		if not len(self.stack) == 1:
			raise AssertionError('Did not finish processing: %r' % [t.tag for t in self.stack[1:]])
		elif not len(self.toplevel) == 1:
			raise AssertionError('Not a single toplevel element: %r' % [t.tag for t in self.toplevel])
		return self.toplevel[0]

	# Builder interface

	def start(self, tag, attrib=None):
		element = self.elementfactory(tag, attrib)
		self.stack[-1].append(element)
		self.stack.append(element)

	def end(self, tag):
		element = self.stack.pop()
		if element.tag != tag:
			raise AssertionError('Unmatched %s at end of %s' % (element.tag, tag))

	def text(self, text):
		self.stack[-1].append(text)

	def append(self, tag, attrib=None, text=None):
		element = self.elementfactory(tag, attrib)
		if text:
			element.append(text)
		self.stack[-1].append(element)


class SimpleTreeParser():
	'''Parser that walks a tree structure if L{SimpleTreeElement}s and calls a L{Builder} object'''

	def __init__(self, builder: Builder):
		self.builder = builder

	def parse(self, elt: SimpleTreeElement):
		self.builder.start(elt.tag, elt.attrib)
		for t in elt:
			if isinstance(t, str):
				self.builder.text(t)
			else:
				self.parse(t) # recurs
		self.builder.end(elt.tag)


def tokens_to_simpletree(tokens) -> SimpleTreeElement:
	'''Helper function to convert between tokens and SimpleTree structure'''
	# Could be aciheved by chaining SimpleTreeBuilder and TokenParser
	# but this is more direct implementation.
	root = SimpleTreeElement(*tokens[0])
	stack = [root]
	for t in tokens[1:]:
		if t[0] == TEXT:
			stack[-1].append(t[1])
		elif t[0] == END:
			assert t[1] == stack[-1].tag
			stack.pop()
		else:
			elt =  SimpleTreeElement(*t)
			stack[-1].append(elt)
			stack.append(elt)

	assert len(stack) == 0
	return root


def simpletree_to_tokens(elt: SimpleTreeElement):
	'''Helper function to convert between SimpleTree structure tokens'''
	# Could be aciheved by chaining TokenBuilder and SimpleTreeParser
	# but this is more direct implementation.
	tokens = [(elt.tag, elt.attrib)]
	for t in elt:
		if isinstance(t, str):
			tokens.append((TEXT, t))
		else:
			tokens.extend(simpletree_to_tokens(t)) # recurs
	tokens.append((END, elt.tag))
	return tokens




# Copyright 2008-2022 Jaap Karssenberg <jaap.karssenberg@gmail.com>

import os
import sys
import logging
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gdk

logger = logging.getLogger('zim.gui')


from zim.config import data_file, value_is_coord, ConfigDict, Boolean, ConfigManager
from zim.signals import DelayedCallback, ConnectorMixin

from zim.notebook import Path, Page, LINK_DIR_BACKWARD, PageNotAvailableError
from zim.notebook.index import IndexNotFoundError
from zim.notebook.operations import ongoing_operation
from zim.history import History, HistoryPath

from zim.actions import action, toggle_action, radio_action, radio_option, \
	get_gtk_actiongroup, initialize_actiongroup, \
	PRIMARY_MODIFIER_STRING, PRIMARY_MODIFIER_MASK
from zim.gui.widgets import \
	MenuButton, \
	Window, Dialog, \
	ErrorDialog, FileDialog, ProgressDialog, MessageDialog, QuestionDialog, \
	ScrolledTextView, \
	gtk_popup_at_pointer, \
	TOP, BOTTOM

from zim.gui.navigation import NavigationModel
from zim.gui.uiactions import UIActions
from zim.gui.customtools import CustomToolManager, CustomToolManagerUI
from zim.gui.insertedobjects import InsertedObjectUI

from zim.gui.pageview import PageView
from zim.gui.pageview.editbar import ToolBarEditBarManager
from zim.gui.notebookview import NotebookView

from zim.plugins import ExtensionBase, extendable, PluginManager
from zim.gui.actionextension import ActionExtensionBase, \
	populate_toolbar_with_actions, os_default_headerbar


MENU_ACTIONS = (
	('file_menu', None, _('_File')), # T: Menu title
	('edit_menu', None, _('_Edit')), # T: Menu title
	('view_menu', None, _('_View')), # T: Menu title
	('insert_menu', None, _('_Insert')), # T: Menu title
	('search_menu', None, _('_Search')), # T: Menu title
	('format_menu', None, _('For_mat')), # T: Menu title
	('tools_menu', None, _('_Tools')), # T: Menu title
	('go_menu', None, _('_Go')), # T: Menu title
	('help_menu', None, _('_Help')), # T: Menu title
	('checkbox_menu', None, _('_Checkbox')), # T: Menu title
)
if sys.platform == "darwin":
	# don't use mnemonics on macOS to allow alt-<letter> shortcuts
	MENU_ACTIONS = tuple((t[0], t[1], t[2].replace('_', '')) for t in MENU_ACTIONS)


#: Preferences for the user interface
ui_preferences = (
	# key, type, category, label, default
	('prefer-dark-theme', 'bool', 'Interface', _('Prefer dark theme') # T: option for preferences dialog
		+ '\n' + _('This option requires a Gtk theme supporting a dark variant') # T: info text for preferences dialog
		+ '\n' + _('This option requires restart of the application'), False), # T: info text for preferences dialog
	('show_headerbar', 'bool', 'Interface', _('Show controls in the window decoration') + '\n' + _('This option requires restart of the application'), os_default_headerbar),
		# T: option for preferences dialog
	('toggle_on_ctrlspace', 'bool', 'Interface', _('Use %s to switch to the side pane') % (PRIMARY_MODIFIER_STRING + '<Space>'), False),
		# T: Option in the preferences dialog - %s will map to either <Control><Space> or <Command><Space> key binding
		# default value is False because this is mapped to switch between
		# char sets in certain international key mappings
	('always_use_last_cursor_pos', 'bool', 'Interface', _('Always use last cursor position when opening a page'), True),
		# T: Option in the preferences dialog
)


def schedule_on_idle(function, args=()):
	'''Helper function to schedule stuff that can be done later, it will
	be triggered on the gtk "idle" signal.

	@param function: function to call
	@param args: positional arguments
	'''
	def callback():
		function(*args)
		return False # delete signal
	GObject.idle_add(callback)


class MainWindowExtension(ActionExtensionBase):
	'''Base class for extending the L{MainWindow}

	Menu actions can be defined by defining an action method
	and specifying where in the menu this should be placed.

	An action method is any object method of the extension method that
	is decorated by the L{action()} or L{toggle_action()} decorators
	(see L{zim.actions}). Such a method is called when the user clicks
	to correcponding menu item or presses the corresponding key binding.
	The decorator is used to define the text to display in the menu
	and the key binding.

	@ivar window: the L{MainWindow}

	@ivar uistate: a L{ConfigDict} to store the extensions ui state or

	The "uistate" is the per notebook state of the interface, it is
	intended for stuff like the last folder opened by the user or the
	size of a dialog after resizing. It is stored in the X{state.conf}
	file in the notebook cache folder. It differs from the preferences,
	which are stored globally and dictate the behavior of the application.
	(To access the preference use C{plugin.preferences}.)
	'''

	def __init__(self, plugin, window):
		'''Constructor
		@param plugin: the plugin object to which this extension belongs
		@param window: the C{Gtk.Window} being extended
		'''
		ExtensionBase.__init__(self, plugin, window)
		self.window = window
		self.uistate = window.notebook.state[plugin.config_key]
		self._add_actions(window.uimanager)
		self.connectto(window, 'destroy')

	def on_destroy(self, window):
		self.destroy()


class WindowBaseMixin(ConnectorMixin, object):
	'''Common logic between MainWindow and PageWindow'''

	def _init_fullscreen_headerbar(self):
		# Create eventbox for headerbar
		self._in_fullscreen_eventbox = False
		self._fullscreen_eventbox = Gtk.EventBox()
		self._fullscreen_eventbox.set_valign(Gtk.Align.START)
		self._fullscreen_eventbox.set_size_request(1, -1) # 1 pixel sensitive area on top of the screen
		self._zim_window_overlay.add_overlay(self._fullscreen_eventbox)
		self._fullscreen_revealer = Gtk.Revealer()
		self._fullscreen_eventbox.add(self._fullscreen_revealer)
		self._fullscreen_headerbar = Gtk.HeaderBar()
		self._fullscreen_revealer.add(self._fullscreen_headerbar)

		close_button = Gtk.Button()
		close_button.set_image(Gtk.Image.new_from_icon_name('view-restore-symbolic',  Gtk.IconSize.BUTTON))
		close_button.set_tooltip_text(_('Leave Fullscreen')) # T: button label for fullscreen window header
		close_button.connect('clicked', lambda o: self.toggle_fullscreen(False))
		self._fullscreen_headerbar.pack_end(close_button)

		self._fullscreen_eventbox.show_all()
		self._fullscreen_eventbox.set_no_show_all(True)
		self._fullscreen_eventbox.hide()
		self._fullscreen_eventbox.connect('enter-notify-event', self._on_fullscreen_eventbox_enter)
		self._fullscreen_eventbox.connect('leave-notify-event', self._on_fullscreen_eventbox_leave)

		# Generic state for fullscreen
		self.maximized = False
		self.isfullscreen = False
		self.connect_after('window-state-event', self.__class__.on_window_state_event)

	def on_window_state_event(self, event):
		if bool(event.changed_mask & Gdk.WindowState.MAXIMIZED):
			self.maximized = bool(event.new_window_state & Gdk.WindowState.MAXIMIZED)
			schedule_on_idle(lambda: self.pageview.scroll_cursor_on_screen())

		if bool(event.changed_mask & Gdk.WindowState.FULLSCREEN):
			self.isfullscreen = bool(event.new_window_state & Gdk.WindowState.FULLSCREEN)
			self.toggle_fullscreen.set_active(self.isfullscreen)
			schedule_on_idle(lambda: self.pageview.scroll_cursor_on_screen())

			if self.isfullscreen:
				self._fullscreen_eventbox.show()
			else:
				self._fullscreen_eventbox.hide()

			self.update_toolbar()

	def _on_fullscreen_eventbox_enter(self, *a):
		if self._headerbar is None and self._toolbar and self._toolbar.get_visible():
			# Do not show fullscreen headerbar if headerbar controls are in toolbar *and* visible
			# because headerbar will be redundant. If toolbar is not visible, we need at least "exit fullscreen"
			return

		self._in_fullscreen_eventbox = True
		self._update_fullscreen_revealer()

	def _on_fullscreen_eventbox_leave(self, *a):
		self._in_fullscreen_eventbox = False
		self._update_fullscreen_revealer()

	def _update_fullscreen_revealer(self, *a):
		# keep showing headerbar as long as any popup menu is open
		show = self._in_fullscreen_eventbox or any(
			c.get_active() for c in self._fullscreen_headerbar.get_children() if isinstance(c, Gtk.MenuButton)
		)
		self._fullscreen_revealer.set_reveal_child(show)

	def _populate_headerbars(self):
		for headerbar in (self._headerbar, self._fullscreen_headerbar):
			if headerbar is not None:
				self._populate_headerbar(headerbar)
				headerbar.show_all()

		for c in self._fullscreen_headerbar.get_children():
			if isinstance(c, Gtk.MenuButton):
				c.connect('toggled', self._update_fullscreen_revealer)

	def _init_toolbar(self):
		# One time setup of the toolbar widget
		assert self.pageview, 'Ensure pageview is initalized to let preferences be loaded'

		self._toolbar = Gtk.Toolbar()
		self._toolbar_editbar_manager = ToolBarEditBarManager(self.pageview, self._toolbar)

		self.setup_toolbar()

		def on_extensions_changed(o, obj):
			if obj in (self, self.pageview):
				self.update_toolbar()

		self.connectto(PluginManager, 'extensions-changed', on_extensions_changed)

		def on_changed_update(o, *a):
			self.update_toolbar()

		self.connectto(CustomToolManager(), 'changed', on_changed_update)
		self.connectto(self.pageview.preferences, 'changed', on_changed_update)

	def setup_toolbar(self, show=None, position=None):
		# Default setup for toolbar - can be called multiple times - also used by "toolbar" plugin
		try:
			self.remove(self._toolbar)
		except ValueError:
			pass

		# Defaults
		if show is None:
			show = not self.preferences['show_headerbar']

		position = TOP if position is None else position

		# Set toolbar in window
		if show:
			self.add_bar(self._toolbar, position=position)
			if position in (TOP, BOTTOM):
				self._toolbar.set_orientation(Gtk.Orientation.HORIZONTAL)
			else: # LEFT, RIGHT
				self._toolbar.set_orientation(Gtk.Orientation.VERTICAL)
			self._toolbar.show()
			self.update_toolbar()
		else:
			self._toolbar.hide()

		return self._toolbar

	def update_toolbar(self):
		# This method updates the toolbar content. See setup_toolbar() to
		# control placement and visibility.
		if self._toolbar.get_visible():
			for item in self._toolbar.get_children():
				self._toolbar.remove(item)
			self._populate_toolbar(self._toolbar)

			if self.isfullscreen and not self.preferences['show_headerbar']:
				close_button = Gtk.ToolButton()
				close_button.set_icon_name('view-restore-symbolic')
				close_button.set_tooltip_text(_('Leave Fullscreen')) # T: button label for fullscreen window header
				close_button.connect('clicked', lambda o: self.toggle_fullscreen(False))
				self._toolbar.insert(close_button, -1)

			self._toolbar.show_all()

	def _populate_toolbar_inner(self, toolbar):
		if not self.pageview.preferences['show_edit_bar']:
			self._toolbar_editbar_manager.populate_toolbar(toolbar)

		populate_toolbar_with_actions(
			self._toolbar, self, self.pageview,
			include_headercontrols=(not self.preferences['show_headerbar']),
			include_customtools=True
		)

	def set_title(self, text):
		Gtk.Window.set_title(self, text)
		if self._headerbar is not None:
			self._headerbar.set_title(text)
		self._fullscreen_headerbar.set_title(text)

	@toggle_action(_('_Fullscreen'), 'F11', icon='view-fullscreen-symbolic', init=False) # T: Menu item
	def toggle_fullscreen(self, show):
		'''Menu action to toggle the fullscreen state of the window'''
		if show:
			self.fullscreen()
		else:
			self.unfullscreen()

	@toggle_action(_('Toggle _Editable'), icon='document-edit-symbolic', init=True, tooltip=_('Toggle editable')) # T: menu item
	def toggle_editable(self, editable):
		'''Menu action to toggle the read-only state of the application
		@emits: readonly-changed
		'''
		readonly = not editable
		if readonly and self.page and self.page.modified:
			# Save any modification now - will not be allowed after switch
			self.pageview.save_changes()

		for group in self.uimanager.get_action_groups():
			for action in group.list_actions():
				if hasattr(action, 'zim_readonly') \
				and not action.zim_readonly:
					action.set_sensitive(not readonly)

		try:
			self.uistate['readonly'] = readonly
		except KeyError:
			pass
		self.emit('readonly-changed', readonly)

	def set_toggle_editable_state(self, editable_uistate):
		'''Set sensitivity of the "toggle_editable" action
		@param editable_uistate: default state if control is sensitive
		'''
		if self.notebook.readonly or self.page.readonly:
			if self.toggle_editable.get_sensitive():
				self.toggle_editable.set_sensitive(False)
				self._set_tooltip_hack(_('Page is read-only and cannot be edited')) # T: message in toggle editable tooltip

		else:
			if not self.toggle_editable.get_sensitive():
				self.toggle_editable.set_sensitive(True)
				self._set_tooltip_hack(self.toggle_editable.tooltip) # reset to default

			self.toggle_editable(editable_uistate)

	def _set_tooltip_hack(self, text):
		for proxy in self.toggle_editable._proxies: # XXX
			if hasattr(proxy, 'set_tooltip_text'):
				proxy.set_tooltip_text(text)

	def _style_toggle_editable_button(self, button):
		def _change_style_on_toggle(button):
			context = button.get_style_context()
			if button.get_active():
				context.remove_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)
			else:
				context.add_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)
		_change_style_on_toggle(button)
		button.connect('toggled', _change_style_on_toggle)

	def do_pane_state_changed(self, pane, *a):
		if not hasattr(self, 'actiongroup') \
		or self._block_toggle_panes:
			return

		action = self.actiongroup.get_action('toggle_panes')
		visible = bool(self.get_visible_panes())
		if visible != action.get_active():
			action.set_active(visible)

	@toggle_action(_('_Side Panes'), 'F9', icon='gtk-index', init=True) # T: Menu item
	def toggle_panes(self, show):
		'''Menu action to toggle the visibility of the all panes
		@param show: when C{True} or C{False} force the visibility,
		when C{None} toggle based on current state
		'''
		self._block_toggle_panes = True
		Window.toggle_panes(self, show)
		self._block_toggle_panes = False

		if show:
			self.focus_sidepane()
		else:
			self.pageview.grab_focus()

		self._sidepane_autoclose = False
		self.save_uistate()

	def do_set_focus(self, widget):
		Window.do_set_focus(self, widget)
		if widget == self.pageview.textview \
		and self._sidepane_autoclose:
			# Sidepane open and should close automatically
			self.toggle_panes(False)

	@action(_('Focus Sidepane')) # T: menu item
	def focus_sidepane_key_toggle(self):
		self.do_focus_sidepane_key_toggle()

	def do_focus_sidepane_key_toggle(self, *a):
		'''Switch focus between the textview and the page index.
		Automatically opens the sidepane if it is closed
		(but sets a property to automatically close it again).
		This method is used for the (optional) <Primary><Space> keybinding.
		'''
		action = self.actiongroup.get_action('toggle_panes')
		if action.get_active():
			# side pane open
			if self.pageview.textview.is_focus():
				self.focus_sidepane()
			else:
				if self._sidepane_autoclose:
					self.toggle_panes(False)
				else:
					self.pageview.grab_focus()
		else:
			# open the pane
			self.toggle_panes(True)
			self._sidepane_autoclose = True

		return True # stop


@extendable(MainWindowExtension)
class MainWindow(WindowBaseMixin, Window):

	# define signals we want to use - (closure type, return type and arg types)
	__gsignals__ = {
		'init-uistate': (GObject.SignalFlags.RUN_LAST, None, ()),
		'page-changed': (GObject.SignalFlags.RUN_LAST, None, (object,)),
		'readonly-changed': (GObject.SignalFlags.RUN_LAST, None, (bool,)),
		'close': (GObject.SignalFlags.RUN_LAST, None, ()),
	}

	def __init__(self, notebook, page=None, fullscreen=False, geometry=None):
		'''Constructor
		@param notebook: the L{Notebook} to show in this window
		@param page: a C{Path} object to open
		@param fullscreen: if C{True} the window is shown fullscreen,
		if C{None} the previous state is restored
		@param geometry: the window geometry as string in format
		"C{WxH+X+Y}", if C{None} the previous state is restored
		'''
		Window.__init__(self)
		self.notebook = notebook
		self.page = None # will be set later by open_page
		self.navigation = NavigationModel(self)
		self.hideonclose = False

		self.preferences = ConfigManager.preferences['GtkInterface']
		self.preferences.define({p[0]: Boolean(p[-1]) for p in ui_preferences})
		self.preferences.connect('changed', self.do_preferences_changed)

		# Hidden setting to force the gtk bell off. Otherwise it
		# can bell every time you reach the begin or end of the text
		# buffer. Especially specific gtk version on windows.
		# See bug lp:546920
		self.preferences.setdefault('gtk_bell', False)
		if not self.preferences['gtk_bell']:
			Gtk.rc_parse_string('gtk-error-bell = 0')

		self._block_toggle_panes = False
		self._sidepane_autoclose = False
		self._switch_focus_accelgroup = None

		# Catching this signal prevents the window to actually be destroyed
		# when the user tries to close it. The action for close should either
		# hide or destroy the window.
		def do_delete_event(*a):
			logger.debug('Action: close (delete-event)')
			self.close()
			return True # Do not destroy - let close() handle it
		self.connect('delete-event', do_delete_event)

		# setup uistate
		self.uistate = notebook.state['MainWindow']
		self.uistate.setdefault('windowpos', None, check=value_is_coord)
		self.uistate.setdefault('windowsize', (600, 450), check=value_is_coord)
		self.uistate.setdefault('windowmaximized', False)
		self.uistate.setdefault('active_tabs', None, tuple)
		self.uistate.setdefault('readonly', False)

		self.history = History(notebook, notebook.state)

		# init uimanager
		self.uimanager = Gtk.UIManager()
		self.uimanager.add_ui_from_string('''
		<ui>
			<menubar name="menubar">
			</menubar>
		</ui>
		''')

		# setup menubar
		self.add_accel_group(self.uimanager.get_accel_group())
		self.menubar = self.uimanager.get_widget('/menubar')
		self.add_bar(self.menubar, position=TOP)

		self.pageview = NotebookView(self.notebook, self.navigation)
		self.connect_object('readonly-changed', NotebookView.set_readonly, self.pageview)

		self.add(self.pageview)

		self.do_preferences_changed()

		self._geometry_set = False
		self._set_fullscreen = False
		if geometry:
			try:
				self.parse_geometry(geometry)
				self._geometry_set = True
			except:
				logger.exception('Parsing geometry string failed:')
		elif fullscreen:
			self._set_fullscreen = True

		# Init mouse settings
		self.preferences.setdefault('mouse_nav_button_back', 8)
		self.preferences.setdefault('mouse_nav_button_forw', 9)

		# Finish uimanager
		self._uiactions = UIActions(self, self.notebook, self.page, self.navigation)
		self.__zim_extension_objects__.append(self._uiactions) # HACK to make actions discoverable
		group = get_gtk_actiongroup(self._uiactions)
		self.uimanager.insert_action_group(group, 0)

		group = get_gtk_actiongroup(self.pageview)
		self.uimanager.insert_action_group(group, 0)

		group = get_gtk_actiongroup(self)
		group.add_actions(MENU_ACTIONS)
		self.uimanager.insert_action_group(group, 0)

		self.open_page_back.set_sensitive(False)
		self.open_page_forward.set_sensitive(False)

		fname = 'menubar.xml'
		self.uimanager.add_ui_from_string(data_file(fname).read())

		# header & tool bars
		if self.preferences['show_headerbar']:
			self._headerbar = Gtk.HeaderBar()
			self._headerbar.set_show_close_button(True)
			self.set_titlebar(self._headerbar)
		else:
			self._headerbar = None

		self._init_fullscreen_headerbar()
		self._populate_headerbars()
		self._init_toolbar()

		# Do this last, else menu items show up in wrong place
		self._customtools = CustomToolManagerUI(self.uimanager, self.pageview)
		self._insertedobjects = InsertedObjectUI(self.uimanager, self.pageview)
			# XXX: would like to do this in PageView itself, but need access to uimanager

		# Setup notebook signals
		notebook.connect('page-info-changed', self.do_page_info_changed)

		def move_away(o, path):
			# Try several options to get awaay
			actions = [self.open_page_back, self.open_page_parent, self.open_page_home]
			while (path == self.page or self.page.ischild(path)) and actions:
				action = actions.pop(0)
				action()

		notebook.connect('deleted-page', move_away) # after action

		def follow(o, path, newpath):
			if path == self.page:
				self.open_page(newpath)
			elif self.page.ischild(path):
				newpath = newpath + self.page.relname(path)
				self.open_page(newpath)
			else:
				pass

		notebook.connect('moved-page', follow) # after action

		# init page
		page = page or self.history.get_current()
		if page:
			page = notebook.get_page(page)
			self.open_page(page)
		else:
			self.open_page_home()

		PluginManager.register_new_extendable(self.pageview)
		initialize_actiongroup(self, 'win')

		self.pageview.grab_focus()

	def _populate_headerbar(self, headerbar):
		hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
		for action in (
			self.open_page_back,
			self.open_page_home,
			self.open_page_forward,
		):
			hbox.add(action.create_icon_button())
		context = hbox.get_style_context()
		context.add_class("linked")
		headerbar.pack_start(hbox)

		headerbar.pack_end(self._uiactions.show_search.create_icon_button())

		button = self.toggle_editable.create_icon_button()
		self._style_toggle_editable_button(button)
		headerbar.pack_end(button)

	def _populate_toolbar(self, toolbar):
		# Add default controls
		if not self.preferences['show_headerbar']:
			for action in (
				self.open_page_back,
				self.open_page_home,
				self.open_page_forward,
			):
				toolbar.insert(action.create_tool_button(), -1)
			toolbar.insert(Gtk.SeparatorToolItem(), -1)

			item = self.toggle_editable.create_tool_button(connect_button=False)
			item.set_action_name('win.toggle_editable')
			self._style_toggle_editable_button(item)
			toolbar.insert(item, -1)
			toolbar.insert(Gtk.SeparatorToolItem(), -1)

			self._populate_toolbar_inner(toolbar)

			space = Gtk.SeparatorToolItem()
			space.set_draw(False)
			space.set_expand(True)
			toolbar.insert(space, -1)

			toolbar.insert(self._uiactions.show_search.create_tool_button(), -1)
		else:
			self._populate_toolbar_inner(toolbar)

	@action(_('_Close'), '<Primary>W') # T: Menu item
	def close(self):
		'''Menu action for close. Will hide when L{hideonclose} is set,
		otherwise destroys window, which could result in the application
		closing if there are no other toplevel windows.
		'''
		if self.hideonclose: # XXX
			self._do_close()
		else:
			self.destroy()

	def _do_close(self):
		self.save_uistate()
		self.hide()
		self.emit('close')

	def destroy(self):
		self.pageview.save_changes()
		if self.page.modified:
			return # Do not quit if page not saved

		self._do_close()

		while Gtk.events_pending():
			Gtk.main_iteration_do(False)

		self.notebook.index.stop_background_check()
		op = ongoing_operation(self.notebook)
		if op:
			op.wait()

		Window.destroy(self) # gtk destroy & will also emit destroy signal

	def do_preferences_changed(self, *a):
		if self._switch_focus_accelgroup:
			self.remove_accel_group(self._switch_focus_accelgroup)

		space = Gdk.unicode_to_keyval(ord(' '))
		group = Gtk.AccelGroup()

		self.preferences.setdefault('toggle_on_altspace', False)
		if self.preferences['toggle_on_altspace']:
			# Hidden param, disabled because it causes problems with
			# several international layouts (space mistaken for alt-space,
			# see bug lp:620315)
			group.connect( # <Alt><Space>
				space, Gdk.ModifierType.MOD1_MASK, Gtk.AccelFlags.VISIBLE,
				self.do_focus_sidepane_key_toggle)

		# Toggled by preference menu, also causes issues with international
		# layouts - esp. when switching input method on Meta-Space
		if self.preferences['toggle_on_ctrlspace']:
			group.connect( # <Primary><Space>
				space, PRIMARY_MODIFIER_MASK, Gtk.AccelFlags.VISIBLE,
				self.do_focus_sidepane_key_toggle)

		self.add_accel_group(group)
		self._switch_focus_accelgroup = group

		# Toggle dark theme
		gtk_settings = Gtk.Settings.get_default()
		text_style = ConfigManager.get_config_dict('style.conf')
		if self.preferences['prefer-dark-theme']:
			gtk_settings.set_property('gtk-application-prefer-dark-theme', True)
			text_style.set_selectors(('darktheme',))
		else:
			gtk_settings.set_property('gtk-application-prefer-dark-theme', False)
			text_style.set_selectors(None)

	@toggle_action(_('Menubar'), init=True) # T: label for View->Menubar menu item
	def toggle_menubar(self, show):
		'''Menu action to toggle the visibility of the menu bar
		@param show: when C{True} or C{False} force the visibility,
		when C{None} toggle based on current state
		'''
		if show:
			self.menubar.set_no_show_all(False)
			self.menubar.show()
		else:
			self.menubar.hide()
			self.menubar.set_no_show_all(True)

	def init_uistate(self):
		# Initialize all the uistate parameters
		# delayed till show or show_all because all this needs real
		# uistate to be in place and plugins to be loaded
		# Run between loading plugins and actually presenting the window to the user

		if not self._geometry_set:
			# Ignore this if an explicit geometry was specified to the constructor
			if self.uistate['windowpos'] is not None:
				x, y = self.uistate['windowpos']
				self.move(x, y)

			w, h = self.uistate['windowsize']
			self.set_default_size(w, h)

			if self.uistate['windowmaximized']:
				self.maximize()

		Window.init_uistate(self) # takes care of sidepane positions etc

		self.toggle_fullscreen(self._set_fullscreen)

		# And hook to notebook properties
		self.on_notebook_properties_changed(self.notebook.properties)
		self.notebook.properties.connect('changed', self.on_notebook_properties_changed)

		# Notify plugins
		self.emit('init-uistate')

		# Update menus etc.
		self.uimanager.ensure_update()
			# Prevent flashing when the toolbar is loaded after showing the window
			# and do this before connecting signal below for accelmap.

		# Load accelmap config and setup saving it
		# TODO - this probably belongs in the application class, not here
		accelmap = ConfigManager.get_config_file('accelmap').file
		logger.debug('Accelmap: %s', accelmap.path)
		if accelmap.exists():
			Gtk.AccelMap.load(accelmap.path)

		def on_accel_map_changed(o, path, key, mod):
			logger.info('Accelerator changed for %s', path)
			Gtk.AccelMap.save(accelmap.path)

		Gtk.AccelMap.get().connect('changed', on_accel_map_changed)

	def save_uistate(self):
		if not self.pageview._zim_extendable_registered:
			return
			# Not allowed to save before plugins are loaded, could overwrite
			# pane state based on empty panes

		cursor = self.pageview.get_cursor_pos()
		scroll = self.pageview.get_scroll_pos()
		self.history.set_state(self.page, cursor, scroll)

		if self.is_visible() and not self.isfullscreen:
			self.uistate['windowpos'] = tuple(self.get_position())
			self.uistate['windowsize'] = tuple(self.get_size())
			self.uistate['windowmaximized'] = self.maximized

		Window.save_uistate(self) # takes care of sidepane positions etc.

		if self.notebook.state.modified:
			self.notebook.state.write()

	def on_notebook_properties_changed(self, properties):
		self._update_window_title()
		if self.notebook.icon:
			try:
				self.set_icon_from_file(self.notebook.icon)
			except (GObject.GError, GLib.Error):
				logger.exception('Could not load icon %s', self.notebook.icon)

	def do_button_press_event(self, event):
		## Try to capture buttons for navigation
		if event.button > 3:
			if event.button == self.preferences['mouse_nav_button_back']:
				self.open_page_back()
			elif event.button == self.preferences['mouse_nav_button_forw']:
				self.open_page_forward()
			else:
				logger.debug("Unused mouse button %i", event.button)
		#~ return Window.do_button_press_event(self, event)

	def open_page(self, path, anchor=None, anchor_fail_silent=False):
		'''Method to open a page in the mainwindow, and menu action for
		the "jump to" menu item.

		Fails silently when saving current page failed (which is usually the
		result of pressing "cancel" in the error dialog shown when saving
		fails). Check return value for success if you want to be sure.

		@param path: a L{path} for the page to open.
		@param anchor: name of an anchor (optional)
		@raises PageNotFound: if C{path} can not be opened
		@emits: page-changed
		@returns: C{True} for success
		'''
		assert isinstance(path, Path)
		try:
			page = self.notebook.get_page(path) # can raise
		except PageNotAvailableError as error:
			# Same code in NewPageDialog
			if QuestionDialog(self, (
				_('File exists, do you want to import?'), # T: short question on open-page if file exists
				_('The file "%s" exists but is not a wiki page.\nDo you want to import it?') % error.file.basename # T: longer question on open-page if file exists
			)).run():
				from zim.import_files import import_file
				page = import_file(error.file, self.notebook, path)
			else:
				return # user cancelled

		if self.page and id(self.page) == id(page):
			if anchor:
				self.pageview.navigate_to_anchor(anchor, fail_silent=anchor_fail_silent)
			return
		elif self.page:
			self.pageview.save_changes() # XXX - should connect to signal instead of call here
			self.notebook.wait_for_store_page_async() # XXX - should not be needed - hide in notebook/page class - how?
			if self.page.modified:
				return False # Assume SavePageErrorDialog was shown and cancelled

			self.save_uistate()

		logger.info('Open page: %s (%s)', page, path)
		self.page = page
		self._uiactions.page = page

		self.notebook.index.touch_current_page_placeholder(path)

		paths = [page] + list(page.parents())
		self.notebook.index.check_async(self.notebook, paths, recursive=False)

		if isinstance(path, HistoryPath):
			self.history.set_current(path)
			cursor = path.cursor # can still be None
		else:
			self.history.append(page)
			cursor = None

		if cursor is None and self.preferences['always_use_last_cursor_pos']:
			cursor, x = self.history.get_state(page)

		self.pageview.set_page(page, cursor)

		if anchor:
			self.pageview.navigate_to_anchor(anchor, fail_silent=anchor_fail_silent)

		self.emit('page-changed', page)

		self.pageview.grab_focus()

	def do_page_changed(self, page):
		self.update_buttons_history()
		self.update_buttons_hierarchy()
		self._update_window_title()
		self.set_toggle_editable_state(not self.uistate['readonly'])

	def _update_window_title(self):
		if self.notebook.readonly or (self.page and self.page.readonly):
			readonly = ' [' + _('readonly') + ']' # T: page status for title bar
		else:
			readonly = ''

		if self.page:
			title = self.page.name + ' - ' + self.notebook.name + readonly
		else:
			title = self.notebook.name + readonly

		self.set_title(title)

	def do_page_info_changed(self, notebook, page):
		if page == self.page:
			self.update_buttons_hierarchy()

	def update_buttons_history(self):
		historyrecord = self.history.get_current()
		self.open_page_back.set_sensitive(not historyrecord.is_first)
		self.open_page_forward.set_sensitive(not historyrecord.is_last)

	def update_buttons_hierarchy(self):
		self.open_page_parent.set_sensitive(len(self.page.namespace) > 0)
		self.open_page_child.set_sensitive(self.page.haschildren)

		has_prev, has_next = self.notebook.pages.get_has_previous_has_next(self.page)
		self.open_page_previous.set_sensitive(has_prev)
		self.open_page_next.set_sensitive(has_next)

	@action(_('_Jump To...'), '<Primary>J') # T: Menu item
	def show_jump_to(self):
		return OpenPageDialog(self, self.page, self.open_page).run()

	@action(
		_('_Back'), verb_icon='go-previous-symbolic', # T: Menu item
		accelerator='<alt>Left', alt_accelerator='XF86Back',
		tooltip=_('Go back') # T: tooltip for navigation button
	)
	def open_page_back(self):
		'''Menu action to open the previous page from the history
		@returns: C{True} if successfull
		'''
		record = self.history.get_previous()
		if not record is None:
			self.open_page(record)

	@action(
		_('_Forward'), verb_icon='go-next-symbolic', # T: Menu item
		accelerator='<alt>Right', alt_accelerator='XF86Forward',
		tooltip=_('Go forward') # T: tooltip for navigation button
	)
	def open_page_forward(self):
		'''Menu action to open the next page from the history
		@returns: C{True} if successfull
		'''
		record = self.history.get_next()
		if not record is None:
			self.open_page(record)

	@action(_('_Parent'), '<alt>Up') # T: Menu item
	def open_page_parent(self):
		'''Menu action to open the parent page
		@returns: C{True} if successful
		'''
		namespace = self.page.namespace
		if namespace:
			self.open_page(Path(namespace))

	@action(_('_Child'), '<alt>Down') # T: Menu item
	def open_page_child(self):
		'''Menu action to open a child page. Either takes the last child
		from the history, or the first child.
		@returns: C{True} if successfull
		'''
		path = self.notebook.pages.lookup_by_pagename(self.page)
			# Force refresh "haschildren" ...
		if path.haschildren:
			record = self.history.get_child(path)
			if not record is None:
				self.open_page(record)
			else:
				child = self.notebook.pages.get_next(path)
				self.open_page(child)

	@action(_('_Previous in Index'), accelerator='<alt>Page_Up') # T: Menu item
	def open_page_previous(self):
		'''Menu action to open the previous page from the index
		@returns: C{True} if successfull
		'''
		path = self.notebook.pages.get_previous(self.page)
		if not path is None:
			self.open_page(path)

	@action(_('_Next in Index'), accelerator='<alt>Page_Down') # T: Menu item
	def open_page_next(self):
		'''Menu action to open the next page from the index
		@returns: C{True} if successfull
		'''
		path = self.notebook.pages.get_next(self.page)
		if not path is None:
			self.open_page(path)

	@action(_('_Home'), '<alt>Home', verb_icon='go-home-symbolic', tooltip=_('Go to home page')) # T: Menu item
	def open_page_home(self):
		'''Menu action to open the home page'''
		self.open_page(self.notebook.get_home_page())


class BackLinksMenuButton(MenuButton):

	def __init__(self, notebook, open_page, status_bar_style=False):
		MenuButton.__init__(self, '-backlinks-', Gtk.Menu(), status_bar_style)
		self.notebook = notebook
		self.open_page = open_page
		self.set_sensitive(False)

	def set_page(self, page):
		self.page = page
		try:
			n = self.notebook.links.n_list_links(self.page, LINK_DIR_BACKWARD)
		except IndexNotFoundError:
			n = 0

		self.label.set_text_with_mnemonic(
			ngettext('%i _Backlink', '%i _Backlinks', n) % n)
			# T: Label for button with backlinks in statusbar
		self.set_sensitive(n > 0)

	def popup_menu(self, event=None):
		# Create menu on the fly
		self.menu = Gtk.Menu()
		notebook = self.notebook
		links = list(notebook.links.list_links(self.page, LINK_DIR_BACKWARD))
		if not links:
			return

		links.sort(key=lambda a: a.source.name)
		for link in links:
			item = Gtk.MenuItem.new_with_mnemonic(link.source.name)
			item.connect_object('activate', self.open_page, link.source)
			self.menu.add(item)

		MenuButton.popup_menu(self, event)


class PageWindow(WindowBaseMixin, Window):
	'''Window to show a single page'''

	# define signals we want to use - (closure type, return type and arg types)
	__gsignals__ = {
		'readonly-changed': (GObject.SignalFlags.RUN_LAST, None, (bool,)),
	}

	def __init__(self, notebook, page, navigation, editable=True):
		Window.__init__(self)
		self._block_toggle_panes = False
		self._sidepane_autoclose = False
		self.navigation = navigation
		self.notebook = notebook
		self.page = notebook.get_page(page)

		self.preferences = ConfigManager.preferences['GtkInterface']
		self.preferences.define({p[0]: Boolean(p[-1]) for p in ui_preferences})
		#self.preferences.connect('changed', self.do_preferences_changed)

		if self.preferences['show_headerbar']:
			self._headerbar = Gtk.HeaderBar()
			self._headerbar.set_show_close_button(True)
			self.set_titlebar(self._headerbar)
		else:
			self._headerbar = None

		self._init_fullscreen_headerbar()
		self._populate_headerbars()

		if self.notebook.readonly or self.page.readonly:
			title = page.name + ' [' + _('readonly') + ']' # T: page status for title bar
		else:
			title = page.name
		self.set_title(title)
		#if ui.notebook.icon:
		#	try:
		#		self.set_icon_from_file(ui.notebook.icon)
		#	except (GObject.GError, GLib.Error):
		#		logger.exception('Could not load icon %s', ui.notebook.icon)


		self.uistate = notebook.state['PageWindow']
		self.uistate.setdefault('windowsize', (500, 400), check=value_is_coord)
		w, h = self.uistate['windowsize']
		self.set_default_size(w, h)

		self.pageview = PageView(notebook, navigation)
		self.connect_object('readonly-changed', PageView.set_readonly, self.pageview)
		self.pageview.set_page(self.page)
		self.add(self.pageview)

		# Need UIManager & menubar to make accelerators and plugin actions work
		self.uimanager = Gtk.UIManager()
		self.add_accel_group(self.uimanager.get_accel_group())

		group = get_gtk_actiongroup(self)
		group.add_actions(MENU_ACTIONS)
		self.uimanager.insert_action_group(group, 0)

		group = get_gtk_actiongroup(self.pageview)
		self.uimanager.insert_action_group(group, 0)

		self._uiactions = UIActions(self, self.notebook, self.page, self.navigation)
		group = get_gtk_actiongroup(self._uiactions)
		self.uimanager.insert_action_group(group, 0)

		fname = 'pagewindow_ui.xml'
		self.uimanager.add_ui_from_string(data_file(fname).read())

		self.menubar = self.uimanager.get_widget('/menubar')
		self.add_bar(self.menubar, position=TOP)

		self._init_toolbar()

		# Close window when page is moved or deleted
		def on_notebook_change(o, path, *a):
			if path == self.page or self.page.ischild(path):
				logger.debug('Close PageWindow for %s (page is gone)', self.page)
				self.save_uistate()
				self.destroy()

		notebook.connect('deleted-page', on_notebook_change)
		notebook.connect('moved-page', on_notebook_change)

		# setup state
		self.set_toggle_editable_state(editable)

		# on close window
		def do_delete_event(*a):
			logger.debug('Close PageWindow for %s', self.page)
			self.save_uistate()

		self.connect('delete-event', do_delete_event)

		PluginManager.register_new_extendable(self.pageview)
		initialize_actiongroup(self, 'win')

		self.pageview.grab_focus()

	def _populate_headerbar(self, headerbar):
		#if headerbar is self._headerbar:
		#	headerbar.pack_end(self.toggle_fullscreen.create_icon_button()) # FIXME: should go in menu popover

		button = self.toggle_editable.create_icon_button()
		self._style_toggle_editable_button(button)
		headerbar.pack_end(button)

	def _populate_toolbar(self, toolbar):
		# Add default controls
		if not self.preferences['show_headerbar']:
			item = self.toggle_editable.create_tool_button(connect_button=False)
			item.set_action_name('win.toggle_editable')
			self._style_toggle_editable_button(item)
			toolbar.insert(item, -1)
			toolbar.insert(Gtk.SeparatorToolItem(), -1)

			self._populate_toolbar_inner(toolbar)

			space = Gtk.SeparatorToolItem()
			space.set_draw(False)
			space.set_expand(True)
			toolbar.insert(space, -1)

			# FIXME: should go in menu popover
			#item = self.toggle_fullscreen.create_tool_button(connect_button=False)
			#item.set_action_name('win.toggle_fullscreen')
			#toolbar.insert(item, -1)
		else:
			self._populate_toolbar_inner(toolbar)

	def save_uistate(self):
		if not self.pageview._zim_extendable_registered:
			return
			# Not allowed to save before plugins are loaded, could overwrite
			# pane state based on empty panes
		self.uistate['windowsize'] = tuple(self.get_size())
		Window.save_uistate(self) # takes care of sidepane positions etc.


class OpenPageDialog(Dialog):
	'''Dialog to go to a specific page. Also known as the "Jump to" dialog.
	Prompts for a page name and navigate to that page on 'Ok'.
	'''

	def __init__(self, parent, page, callback):
		Dialog.__init__(self, parent, _('Jump to'), # T: Dialog title
			button=_('_Jump'), # T: Button label
		)
		self.callback = callback

		self.add_form(
			[('page', 'page', _('Jump to Page'), page)], # T: Label for page input
			notebook=parent.notebook
		)

	def do_response_ok(self):
		path = self.form['page']
		if not path:
			return False
		self.callback(path)
		return True




# Copyright 2009 Jaap Karssenberg <jaap.karssenberg@gmail.com>

import logging

from zim.plugins import PluginClass
from zim.plugins.base.imagegenerator import \
	ImageGeneratorClass, BackwardImageGeneratorObjectType

from zim.newfs import LocalFile, TmpFile
from zim.applications import Application, ApplicationError
from zim.gui.base.images import supports_image_format

logger = logging.getLogger('zim.plugins.diagrameditor')

def get_cmd(fmt):
	return ('dot', f'-T{fmt}', '-o')


class InsertDiagramPlugin(PluginClass):

	plugin_info = {
		'name': _('Insert Diagram'), # T: plugin name
		'description': _('''\
This plugin provides a diagram editor for zim based on GraphViz.

This is a core plugin shipping with zim.
'''), # T: plugin description
		'help': 'Plugins:Diagram Editor',
		'author': 'Jaap Karssenberg',
	}

	plugin_preferences = (
		# key, type, label, default
		(
			'prefer_svg',
			'bool',
            _('Generate diagrams in SVG format'), # T: plugin preference
			supports_image_format('svg')
		),
		(
			'default_text',
			'string',
			_('Default text'), # T: Plugin preference for diagram editor
			''
		),
	)

	@classmethod
	def check_dependencies(klass):
		has_dotcmd = Application(get_cmd('png')).tryexec()
		return has_dotcmd, [("GraphViz", has_dotcmd, True)]


class BackwardDiagramImageObjectType(BackwardImageGeneratorObjectType):

	name = 'image+diagram'
	label = _('Diagram') # T: menu item
	syntax = 'dot'
	scriptname = 'diagram.dot'


class DiagramGenerator(ImageGeneratorClass):

	@property
	def _pref_format(self):
		return 'svg' if self.plugin.preferences['prefer_svg'] else 'png'

	@property
	def imagefile_extension(self):
		return '.' + self._pref_format

	@property
	def dotcmd(self):
		return get_cmd(self._pref_format)

	def __init__(self, plugin, notebook, page):
		ImageGeneratorClass.__init__(self, plugin, notebook, page)
		self.dotfile = TmpFile('diagram.dot')
		self.dotfile.touch()
		self.default_text = plugin.preferences['default_text']

	def generate_image(self, text):
		# Write to tmp file
		self.dotfile.write(text)
		self.imgfile = LocalFile(self.dotfile.path[:-4] + self.imagefile_extension) # len('.dot') == 4
		logger.debug('Writing diagram to temp file: %s', self.imgfile)

		# Call GraphViz
		try:
			dot = Application(self.dotcmd)
			dot.run((self.imgfile, self.dotfile))
		except ApplicationError:
			return None, None # Sorry, no log
		else:
			if self.imgfile.exists():
				return self.imgfile, None
			else:
				# When supplying a dot file with a syntax error, the dot command
				# doesn't return an error code (so we don't raise
				# ApplicationError), but we still don't have a png file to
				# return, so return None.
				return None, None

	def cleanup(self):
		self.dotfile.remove()
		try:
			self.imgfile.remove()
		except AttributeError:
			logger.debug('Closed dialog before generating image, nothing to remove')

	def get_default_text(self):
		'''Provides a template or starting point for the user to begin editing.'''
		return self.default_text or ''




# Copyright 2009,2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module contains helper classes for running external applications.

See L{zim.gui.applications} for classes with desktop integration for
applications defined in desktop entry files.
'''

import sys
import os
import re
import logging
import subprocess
import locale

from gi.repository import GObject
from gi.repository import GLib

import zim.errors

from zim.fs import adapt_from_oldfs
from zim.newfs import SEP, is_abs_filepath, FilePath, LocalFile
from zim.parse.links import is_uri_re, is_win32_path_re


logger = logging.getLogger('zim.applications')


TEST_MODE = False
TEST_MODE_RUN_CB = None

_ENCODING = locale.getpreferredencoding()

def _decode(data):
	# Since we do not know for sure what encoding other processes will use
	# for output, we need to guess :(
	try:
		return data.decode('UTF-8')
	except UnicodeDecodeError:
		return data.decode(_ENCODING)


_FLATPAK_HOSTCOMMAND_PREFIX = ("flatpak-spawn", "--host")

def _check_flatpak_host_command():
	# Detect whether we are running in Flatpak and can call HostCommand
	if os.path.exists("/.flatpak-info"):
		try:
			subprocess.check_call(_FLATPAK_HOSTCOMMAND_PREFIX + ("which", "which"), stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
			return True
		except subprocess.CalledProcessError:
			pass  # Not privileged to call host command or "which" not found in the host
		except OSError:
			pass  # Failed to execute flatpak-spawn
	return False

_CAN_CALL_FLATPAK_HOST_COMMAND = _check_flatpak_host_command()

def _main_is_frozen():
	# Detect whether we are running py2exe compiled version
	return hasattr(sys, 'frozen') and sys.frozen


def _split_environ_list(value):
	if isinstance(value, str):
		return value.split(os.pathsep)
	elif value is None:
		return []
	else:
		raise ValueError



_word_re = re.compile(r'''
	(
		"(\\"|[^"])*" |  # double quoted word
		[^\s"]+          # word without spaces
	)''', re.X)


def split_quoted_strings(string):
	'''Split a word list respecting quotes according to XDG desktop entry spec

	Only supports double quotes as specified in the spec

	This function always expect full words to be quoted, even if quotes
	appear in the middle of a word, they are considered word
	boundries.

	( XDG Desktop Entry spec says full words must be quoted and
	quotes in a word escaped, but doesn't specify what to do with
	loose quotes in a string. Also this spec does not allow single
	quote quotes)
	'''
	string = string.strip()
	words = []
	m = _word_re.match(string)
	while m:
		w = m.group(0)

		words.append(w)
		i = m.end()
		string = string[i:].lstrip()
		m = _word_re.match(string)

	if string:
		words += string.split() # unmatched quote ?

	return [_unescape_quoted_string(w) for w in words if w]


def _unescape_quoted_string(string):
	# XDG Desktop entry spec says:"If an argument contains a reserved character
	# the argument *must* be quoted."
	# Therefore, unquoted arguments with backslash are invalid. However on
	# Windows we may have created these, so for backward compatibility pass
	# them through without unescaping.
	# Unescaping here does not target \n etc. but \" and other reserved characters
	# avoid touching alphabetic chars in case there are invalid paths in the string
	if string[0] == '"' and string[-1] == '"':
		return re.sub(r'\\(\W)', '\\1', string[1:-1])
	else:
		return string


class ApplicationLookUpError(zim.errors.Error):
	'''Error raised when an application is not found'''

	description = None

	def __init__(self, cmd):
		self.msg = _('Cound not find application: %s') % cmd
			# T: Error message when external application could not be found, %s is the command


class ApplicationError(zim.errors.Error):
	'''Error raised for errors in the sub process'''

	description = None

	def __init__(self, cmd, args, retcode, stderr):
		'''Constructor

		@param cmd: the application command as string
		@param args: tuple of arguments given to the command
		@param retcode: the return code of the command (non-zero!)
		@param stderr: the error output of the command
		'''
		self.msg = _('Failed to run application: %s') % cmd
			# T: Error message when external application failed, %s is the command
		self.description = \
			_('%(cmd)s\nreturned non-zero exit status %(code)i') \
			% {'cmd': cmd + ' "' + '" "'.join(args) + '"', 'code': retcode}
			# T: Error message when external application failed, %(cmd)s is the command, %(code)i the exit code

		if stderr:
			self.description += '\n\n' + stderr


class Application(object):
	'''Base class for objects representing an external application or
	command.

	@ivar name: the name of the command (default to first item of C{cmd})
	@ivar cmd: the command and arguments as a tuple or a string
	(when given as a string it will be parsed for quoted arguments)
	@ivar tryexeccmd: the command to check in L{tryexec()}, if C{None}
	fall back to first item of C{cmd}
	'''

	STATUS_OK = 0 #: return code when the command executed successfullly

	def __init__(self, cmd, tryexeccmd=None):
		'''Constructor

		@param cmd: the command for the external application, either a
		string for the command, or a tuple or list with the command
		and arguments
		@param tryexeccmd: command to check in L{tryexec()} as string.
		If C{None} will default to C{cmd} or the first item of C{cmd}.
		'''
		if isinstance(cmd, str):
			cmd = split_quoted_strings(cmd)
		else:
			assert isinstance(cmd, (tuple, list))
		assert tryexeccmd is None or isinstance(tryexeccmd, str)
		self.cmd = tuple(cmd)
		self.tryexeccmd = tryexeccmd

	def __eq__(self, other):
		if isinstance(other, Application):
			return (self.cmd, self.tryexeccmd) == (other.cmd, other.tryexeccmd)
		else:
			return False

	def __repr__(self):
		if hasattr(self, 'key'):
			return '<%s: %s>' % (self.__class__.__name__, self.key)
		elif hasattr(self, 'cmd'):
			return '<%s: %s>' % (self.__class__.__name__, self.cmd)
		else:
			return '<%s: %s>' % (self.__class__.__name__, self.name)

	@property
	def name(self):
		return self.cmd[0]

	@staticmethod
	def _lookup(cmd):
		'''Lookup cmd in PATH'''
		if is_abs_filepath(cmd):
			if os.path.isfile(cmd):
				return cmd
			else:
				return None
		elif os.name == 'nt':
			# Check executable extensions from windows environment
			extensions = _split_environ_list(os.environ.get('PATHEXT', '.com;.exe;.bat;.cmd'))
			for dir in _split_environ_list(os.environ.get('PATH')):
				for ext in extensions:
					file = SEP.join((dir, cmd + ext))
					if os.path.isfile(file) and os.access(file, os.X_OK):
						return file
			else:
				return None
		else:
			# On POSIX no extension is needed to make scripts executable
			for dir in _split_environ_list(os.environ.get('PATH')):
				file = SEP.join((dir, cmd))
				if os.path.isfile(file) and os.access(file, os.X_OK):
					return file
			else:
				if _CAN_CALL_FLATPAK_HOST_COMMAND:
					try:
						file = subprocess.check_output(_FLATPAK_HOSTCOMMAND_PREFIX + ("which", cmd), stderr=subprocess.DEVNULL)
						return file
					except subprocess.CalledProcessError:
						pass
				return None

	def _cmd(self, args):
		# substitute args in the command - to be overloaded by child classes
		if args:
			return self.cmd + tuple(map(str, args))
		else:
			return self.cmd

	def tryexec(self):
		'''Check if the executable exists without calling it. This
		method is used e.g. to decide what applications to show in the
		gui. Uses the C{tryexeccmd}, or the first item of C{cmd} as the
		executable name.
		@returns: C{True} when the executable was found
		'''
		cmd = self.tryexeccmd or self.cmd[0]
		return not self._lookup(cmd) is None

	def _checkargs(self, cwd, args):
		assert args is None or isinstance(args, (tuple, list))
		argv = self._cmd(args)

		# Expand home dir
		if argv[0].startswith('~'):
			cmd = LocalFile(argv[0]).path
			argv = list(argv)
			argv[0] = cmd

		# if it is a python script, insert interpreter as the executable
		if argv[0].endswith('.py') and not _main_is_frozen():
			argv = list(argv)
			argv.insert(0, sys.executable)
		# TODO: consider an additional commandline arg to re-use compiled python interpreter

		if hasattr(cwd, 'path'):
			cwd = cwd.path

		return cwd, argv

	def run(self, args=None, cwd=None):
		'''Run the application in a sub-process and wait for it to finish.
		Even when the application runs successfully, any message to stderr
		is logged as a warning by zim.
		@param args: additional arguments to give to the command as tuple or list
		@param cwd: the folder to set as working directory for the command
		@raises ApplicationError: if the sub-process returned an error.
		'''
		cwd, argv = self._checkargs(cwd, args)
		logger.info('Running: %r (cwd: %r)', argv, cwd)
		if TEST_MODE:
			TEST_MODE_RUN_CB(argv)
			return None

		if os.name == 'nt':
			# http://code.activestate.com/recipes/409002/
			info = subprocess.STARTUPINFO()
			try:
				info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
			except AttributeError:
				info.dwFlags |= 1 # STARTF_USESHOWWINDOW = 0x01

			p = subprocess.Popen(argv,
				cwd=cwd,
				stdout=subprocess.DEVNULL,
				stderr=subprocess.PIPE,
				startupinfo=info,
				bufsize=4096,
				#~ close_fds=True
			)
		else:
			try:
				p = subprocess.Popen(argv,
					cwd=cwd,
					stdout=subprocess.DEVNULL,
					stderr=subprocess.PIPE,
					bufsize=4096,
					close_fds=True
				)
			except OSError:
				if _CAN_CALL_FLATPAK_HOST_COMMAND:
					p = subprocess.Popen(_FLATPAK_HOSTCOMMAND_PREFIX + argv,
						cwd=cwd,
						stdout=subprocess.DEVNULL,
						stderr=subprocess.PIPE,
						bufsize=4096,
						close_fds=True
					)
				else:
					raise
		stdout, stderr = p.communicate()

		if not p.returncode == self.STATUS_OK:
			raise ApplicationError(argv[0], argv[1:], p.returncode, _decode(stderr))
		#~ elif stderr:
			#~ logger.warn(_decode(stderr))

	def pipe(self, args=None, cwd=None, input=None):
		'''Run the application in a sub-process and capture the output.
		Like L{run()}, but connects to stdin and stdout for the sub-process.

		@note: The data read is buffered in memory, so do not use this
		method if the data size is large or unlimited.

		@param args: additional arguments to give to the command as tuple or list
		@param cwd: the folder to set as working directory for the command
		@param input: input for the command as string

		@returns: output as a list of lines
		@raises ApplicationError: if the sub-process returned an error.
		'''
		cwd, argv = self._checkargs(cwd, args)
		logger.info('Running: %r (cwd: %r)', argv, cwd)
		if TEST_MODE:
			return TEST_MODE_RUN_CB(argv)

		startupinfo = None
		if os.name == 'nt':
			startupinfo = subprocess.STARTUPINFO()
			startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
			startupinfo.wShowWindow = subprocess.SW_HIDE
		try:
			p = subprocess.Popen(argv, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, startupinfo=startupinfo)
		except OSError:
			if _CAN_CALL_FLATPAK_HOST_COMMAND:
				p = subprocess.Popen(_FLATPAK_HOSTCOMMAND_PREFIX + argv, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			else:
				raise
		if input is None:
			stdout, stderr = p.communicate()
		else:
			data = input if isinstance(input, bytes) else str(input).encode('UTF-8')
				# No way to know what encoding the process accepts, so UTF-8 is as good as any
			stdout, stderr = p.communicate(data)

		#~ if not p.returncode == self.STATUS_OK:
			#~ raise ApplicationError(argv[0], argv[1:], p.returncode, stderr)
		#~ elif stderr:
		if stderr:
			logger.warning(_decode(stderr))
			# TODO: allow user to get this error as well - e.g. for logging image generator cmd

		text = _decode(stdout).replace('\r\n', '\n').splitlines(keepends=True)
		return text

	def spawn(self, args=None, callback=None, data=None, cwd=None):
		'''Start the application in the background and return immediately.
		This is used to start an external in parallel with zim that is
		not expected to exit immediatly, so we do not want to wait for
		it - e.g. a webbrowser to show an URL that was clicked.

		@param args: additional arguments to give to the command as tuple or list
		@param callback: optional callback can be used to trigger when
		the application exits. The signature is::

			callback(status, data)

		where 'C{status}' is the exit status of the process. The
		application object provides a constant 'C{STATUS_OK}' which can
		be used to test if the application was successful or not.
		@param data: additional data for the callback
		@param cwd: the folder to set as working directory for the command
		@returns: the PID for the new process
		'''
		cwd, argv = self._checkargs(cwd, args)

		flags = GObject.SPAWN_SEARCH_PATH
		if callback:
			flags |= GObject.SPAWN_DO_NOT_REAP_CHILD
			# without this flag child is reaped automatically -> no zombies

		if cwd is None:
			cwd = os.getcwd()

		logger.info('Spawning: %s (cwd: %s)', argv, cwd)
		if TEST_MODE:
			TEST_MODE_RUN_CB(argv)
			return None

		# https://github.com/zim-desktop-wiki/zim-desktop-wiki/issues/1697
		def _callback_wrapper(pid, *args):
			GLib.spawn_close_pid(pid)
			callback(*args)

		try:
			try:
				pid, stdin, stdout, stderr = \
					GObject.spawn_async(argv, flags=flags, working_directory=cwd)
			except (GObject.GError, GLib.Error):
				if _CAN_CALL_FLATPAK_HOST_COMMAND:
					pid, stdin, stdout, stderr = \
						GObject.spawn_async(_FLATPAK_HOSTCOMMAND_PREFIX + argv, flags=flags, working_directory=cwd)
				else:
					raise
		except (GObject.GError, GLib.Error):
			from zim.gui.widgets import ErrorDialog
			ErrorDialog(None, _('Failed running: %s') % argv[0]).run()
				#~ # T: error when application failed to start
			return None
		else:
			logger.debug('Process started with PID: %i', pid)
			if callback:
				# child watch does implicit reaping -> no zombies
				if data is None:
					GObject.child_watch_add(pid,
						lambda _, status: _callback_wrapper(pid, status))
				else:
					GObject.child_watch_add(pid,
						lambda _, status, data: _callback_wrapper(pid, status, data), data)
			return pid


class WebBrowser(Application):
	'''Application wrapper for the C{webbrowser} module. Can be used as
	fallback if no webbrowser is configured.
	'''

	name = _('Default') + ' (webbrowser)' # T: label for default webbrowser
	key = 'webbrowser' # Used by zim.gui.applications

	def __init__(self):
		import webbrowser
		self.controller = None
		try:
			self.controller = webbrowser.get()
		except webbrowser.Error:
			pass # webbrowser throws an error when no browser is found

	def __eq__(self, other):
		return isinstance(other, self.__class__)

	def tryexec(self):
		return not self.controller is None

	def run(self, args):
		'''This method is not supported by this class
		@raises NotImplementedError: always
		'''
		raise NotImplementedError('WebBrowser can not run in foreground')

	def spawn(self, args, callback=None):
		if callback:
			raise NotImplementedError('WebBrowser can not handle callback')

		for url in args:
			if hasattr(url, 'uri'):
				url = url.uri
			logger.info('Opening in webbrowser: %s', url)

			if TEST_MODE:
				TEST_MODE_RUN_CB((self.__class__.__name__, url))
			else:
				self.controller.open(url)


class StartFile(Application):
	'''Application wrapper for C{os.startfile()}. Can be used on
	windows to open files and URLs with the default application.
	'''

	name = _('Default') + ' (os)' # T: label for default application
	key = 'startfile' # Used by zim.gui.applications

	def __init__(self):
		pass

	def __eq__(self, other):
		return isinstance(other, self.__class__)

	def tryexec(self):
		return hasattr(os, 'startfile')

	def run(self, args):
		'''This method is not supported by this class
		@raises NotImplementedError: always
		'''
		raise NotImplementedError('StartFile can not run in foreground')

	def spawn(self, args, callback=None):
		if callback:
			raise NotImplementedError('os.startfile does not support a callback')

		for arg in args:
			arg = adapt_from_oldfs(arg)
			if hasattr(arg, 'path'):
				path = os.path.normpath(arg.path).replace('/', SEP) # msys can use '/' instead of '\\'
			elif is_uri_re.match(arg) and not is_win32_path_re.match(arg) and not arg.startswith('file://'):
				# URL or e.g. mailto: or outlook: URI
				path = str(arg)
			else:
				# must be file as string
				try:
					arg = FilePath(arg).path
				except ValueError:
					pass

				path = os.path.normpath(str(arg)).replace('/', SEP) # msys can use '/' instead of '\\'

			logger.info('Opening with os.startfile: %s', path)
			if TEST_MODE:
				TEST_MODE_RUN_CB((self.__class__.__name__, path))
			else:
				os.startfile(path)




# Copyright 2008-2017 Jaap Karssenberg <jaap.karssenberg@gmail.com>



from gi.repository import Gtk

import logging

import zim.formats
import zim.templates

from zim.newfs import LocalFile, LocalFolder, TmpFile
from zim.notebook import Path, encode_filename
from zim.notebook.operations import NotebookOperation

from zim.gui.widgets import Assistant, AssistantPage, \
	ProgressDialog, ErrorDialog, QuestionDialog, \
	MessageDialog, LogFileDialog
from zim.gui.applications import open_file

from zim.export import *
from zim.export.selections import *


logger = logging.getLogger('zim.export')


class ExportDialog(Assistant):

	def __init__(self, widget, notebook, page=None):
		Assistant.__init__(self, widget, _('Export'), # T: dialog title
			help=':Help:Export', defaultwindowsize=(400, 325))
		self.notebook = notebook

		self.append_page(InputPage(self, page))
		self.append_page(FormatPage(self))
		self.append_page(OutputPage(self))

	def do_response_ok(self):
		output, exporter = self.get_exporter()
		selection = self.get_selection()
		logger.debug('exporter: %s, selection: %s',
			exporter.__class__.__name__,
			selection.__class__.__name__)
		if exporter is None or selection is None:
			logger.debug('Cancelled - selection')
			return False # canceled

		# Run export
		logging_context = LogContext()
		with logging_context:
			op = NotebookOperation(
				self.notebook,
				_('Exporting notebook'), # T: Title for progressbar window
				exporter.export_iter(selection)
			)
			dialog = ProgressDialog(self, op)
			dialog.run()

		if op.exception:
			raise op.exception
		if op.cancelled:
			return False
		else:
			#~ print('>>> %s E: %i, W: %i' % ()
				#~ logging_context.file.path,
				#~ logging_context.handler.n_error, logging_context.handler.n_warning)
			#~ print logging_context.file.read()
			#~ print('---')
			ExportDoneDialog(self, logging_context, output).run()
			return True

	def get_selection(self):
		if self.uistate['selection'] == 'all':
			return AllPages(self.notebook)
		else:
			path = self.uistate['selected_page']
			if self.uistate['selection_recursive']:
				return SubPages(self.notebook, path)
			else:
				return SinglePage(self.notebook, path)

	def get_exporter(self):
		#~ import pprint
		#~ pprint.pprint(self.uistate)
		#~ return True

		options = {}
		for k in ('format', 'template', 'index_page'):
			if self.uistate[k] and not self.uistate[k].isspace():
				options[k] = self.uistate[k]

		options['format'] = \
			zim.formats.canonical_name(options['format'])

		if options['template'] == '__file__':
			options['template'] = self.uistate['template_file']

		if self.uistate['document_root'] == 'url':
			options['document_root_url'] = self.uistate['document_root_url']

		if options['format'] == 'mhtml':
			# Export MHTML file
			options.pop('format', None)
			options.pop('index_page', None)
			file = self.get_file()
			if file:
				return file, build_mhtml_file_exporter(file, **options)
			else:
				return None, None
		elif self.uistate['output'] == 'single_file':
			options.pop('index_page', None)
			if self.uistate['selection'] == 'page':
				options['namespace'] = self.uistate['selected_page']

			file = self.get_file()
			if file:
				return file, build_single_file_exporter(file, **options)
			else:
				return None, None
		elif self.uistate['selection'] == 'all':
			# Export full notebook to dir
			dir = self.get_folder()
			if dir:
				return dir, build_notebook_exporter(dir, **options)
			else:
				return None, None
		else:
			# Export page to file
			options.pop('index_page', None)
			file = self.get_file()
			if file:
				path = self.uistate['selected_page']
				return file, build_page_exporter(file, page=path, **options)
			else:
				return None, None

	def get_folder(self):
		dir = LocalFolder(self.uistate['output_folder'])
		if dir.exists() and len(dir.list_names()) > 0:
			ok = QuestionDialog(self, (
				_('Folder exists: %s') % dir.path, # T: message heading
				_('Folder already exists and has content, '
				  'exporting to this folder may overwrite '
				  'existing files. '
				  'Do you want to continue?') # T: detailed message, answers are Yes and No
			)).run()
			if not ok:
				return None
		return dir

	def get_file(self):
		file = LocalFile(self.uistate['output_file'])
		if file.exists():
			ok = QuestionDialog(self, (
				_('File exists'), # T: message heading
				_('This file already exists.\n'
				  'Do you want to overwrite it?') # T: detailed message, answers are Yes and No
			)).run()
			if not ok:
				return None
		return file


class InputPage(AssistantPage):
	'''Assistant page allowing to select the page(s) for export'''

	title = _('Select the pages to export') # T: title of step in export dialog

	def __init__(self, assistant, page=None):
		AssistantPage.__init__(self, assistant)

		self.add_form((
			('selection:all', 'option', _('Complete _notebook')),
				# T: Option in export dialog to export complete notebook
			#~ ('selection:selection', 'option', _('_Selection')),
				# T: Option in export dialog to export selection
			#~ ('selection_query', 'string', None),
			('selection:page', 'option', _('Single _page')),
				# T: Option in export dialog to export selection
			None,
			('page', 'page', _('Page')), # T: Input field in export dialog
			('recursive', 'bool', _('Include subpages')), # T: Input field in export dialog
		), {
			'page': page,
			'recursive': True,
		},
		depends={
			'page': 'selection:page',
			'recursive': 'selection:page',
		})
		self.form.widgets['page'].existing_only = True

	def init_uistate(self):
		#~ self.uistate.setdefault('selection', 'all', ('all', 'page'))
		self.uistate.setdefault('selection', 'all')
		self.uistate.setdefault('selection_recursive', True)
		self.uistate.setdefault('selected_page', self.form['page'], check=Path)
		self.form['selection'] = self.uistate['selection']

	def save_uistate(self):
		self.uistate['selection'] = self.form['selection']
		self.uistate['selected_page'] = self.form['page']
		self.uistate['selection_recursive'] = self.form['recursive']


class FormatPage(AssistantPage):
	'''Assistant page allowing to select the output format and template'''

	title = _('Select the export format') # T: title of step in export dialog

	CHOICE_OTHER = _('Other...')
		# T: Option in drop down menu to specify another file

	# FUTURE: this page could have a dynamic form with additional
	# fields requested by the template, e.g. author etc.

	def __init__(self, assistant):
		AssistantPage.__init__(self, assistant)
		self.export_formats = zim.formats.list_formats(zim.formats.EXPORT_FORMAT)
		self.export_formats.insert(1, 'MHTML (Web Page Archive)') # TODO translatable

		self.add_form((
			('format', 'choice', _('Format'), self.export_formats), # T: Input label in the export dialog
			('template', 'choice', _('Template'), ()), # T: Input label in the export dialog
			('template_file', 'file', None),
			None,
			('document_root:absolute', 'option', _('Link files under document root with full file path')), # T: radio option in export dialog
			('document_root:url', 'option', _('Map document root to URL') + ': '), # T: radio option in export dialog
			('document_root_url', 'string', None),
		), depends={
			'document_root_url': 'document_root:url'
		})

		## Same button appears in edit preferences dialog
		url_button = Gtk.LinkButton(
			'https://github.com/jaap-karssenberg/zim-wiki/wiki/Templates',
			_('Get more templates online') # T: label for button with URL
		)
		self.pack_start(url_button, False, True, 0)


		# Set template list based on selected format
		def set_templates(self):
			format = self.form['format']
			format = zim.formats.canonical_name(format)
			if format == 'mhtml':
				format = 'html'
			combobox = self.form.widgets['template']
			combobox.get_model().clear()

			for name, _ in zim.templates.list_templates(format):
				combobox.append_text(name)
			combobox.append_text(self.CHOICE_OTHER)
			combobox.set_sensitive(True)

			template = self.uistate['template']
			try:
				if template == '__file__':
					self.form['template'] = self.CHOICE_OTHER
				else:
					self.form['template'] = template
			except ValueError:
				combobox.set_active(0)

		self.form.widgets['format'].connect_object('changed', set_templates, self)

		# Hook template entry to be sensitive on "Other.."
		self.form.widgets['template_file'].set_sensitive(False)
		self.form.widgets['template'].connect('changed',
			lambda o: self.form.widgets['template_file'].set_sensitive(
							o.get_active_text() == self.CHOICE_OTHER))

		# Check if we have a document root - if not disable all options
		docroot = assistant.notebook.document_root
		if not docroot:
			for widget in self.form.widgets:
				if widget.startswith('document_root:'):
					self.form.widgets[widget].set_sensitive(False)
			self.uistate.input(document_root_url='')

	def init_uistate(self):
		self.uistate.setdefault('format', 'HTML')
		self.uistate.setdefault('template', 'Default')
		self.uistate.setdefault('template_file', '')
		self.uistate.setdefault('document_root', 'absolute', check={'absolute', 'url'})
		self.uistate.setdefault('document_root_url', '')

		try:
			self.form['format'] = self.uistate['format']
			if self.uistate['template'] == '__file__':
				self.form['template'] = self.CHOICE_OTHER
			else:
				self.form['template'] = self.uistate['template']
		except ValueError:
			pass

		self.form['template_file'] = self.uistate['template_file']
		self.form['document_root'] = self.uistate['document_root']
		self.form['document_root_url'] = self.uistate['document_root_url']

	def save_uistate(self):
		self.uistate.update(self.form)
		if self.uistate['template'] == self.CHOICE_OTHER:
			self.uistate['template'] = '__file__'


class OutputPage(AssistantPage):
	'''Assistant page allowing to select output file or folder'''

	title = _('Select the output file or folder') # T: title of step in export dialog

	def __init__(self, assistant):
		AssistantPage.__init__(self, assistant)

		self.add_form((
			('output:multi_file', 'option', _('Export each page to a separate file')),
				# T: Label for option in export dialog
			('output:single_file', 'option', _('Export all pages to a single file')),
				# T: Label for option in export dialog
			None,
			('folder', 'dir', _('Output folder')),
				# T: Label for folder selection in export dialog
			('index', 'string', _('Index page')),
				# T: Label for setting a name for the index of exported pages
				# TODO validation for this entry - valid name, but not existing
			('file', 'output-file', _('Output file')),
				# T: Label for file selection in export dialog
		), depends={
			'index': 'folder',
		})

		for widget in self.form.widgets:
			self.form.widgets[widget].set_no_show_all(True)

		self.form.widgets['output:single_file'].connect('toggled', self.on_output_changed)

	def init_uistate(self):
		# Switch between folder selection or file selection based
		# on whether we selected full notebook or single page in the
		# first page
		self.uistate.setdefault('output', 'multi_file')
		self.uistate.setdefault('output_folder', None, LocalFolder)
		self.uistate.setdefault('index_page', '')
		self.uistate.setdefault('output_file', None, LocalFile)

		if self.uistate.get('format', '').startswith('MHTML'):
			# XXX make this a format property to be queried
			self.form.widgets['output:multi_file'].set_sensitive(False)
			self.form.widgets['output:single_file'].set_sensitive(False)
			self.form.widgets['output:single_file'].set_active(True)
		else:
			self.form.widgets['output:multi_file'].set_sensitive(True)
			self.form.widgets['output:single_file'].set_sensitive(True)

		self.form.widgets['output:multi_file'].show()
		self.form.widgets['output:single_file'].show()

		self.form['output'] = self.uistate['output']

		self.on_output_changed(None)

		# Set file name
		basename = self.uistate['selected_page'].basename
		format = self.uistate['format']
		if format.startswith('MHTML'):
			ext = 'mht'
		else:
			ext = zim.formats.get_format(format).info['extension']

		if self.uistate['output_file'] \
		and isinstance(self.uistate['output_file'], LocalFile):
			dir = self.uistate['output_file'].parent()
			file = dir.file(encode_filename(basename + '.' + ext))
		else:
			file = LocalFile('~/' + encode_filename(basename + '.' + ext))
		self.uistate['output_file'] = file

		self.form['file'] = self.uistate['output_file']
		self.form['folder'] = self.uistate['output_folder']

	def on_output_changed(self, o):
		if self.uistate.get('selection') == 'page':
			self.set_show_file(True)
		else:
			self.set_show_file(self.form['output'] == 'single_file')

	def set_show_file(self, show_file):
		if show_file:
			self.form.widgets['folder'].set_sensitive(False)
			self.form.widgets['folder'].hide()
			self.form.widgets['file'].set_sensitive(True)
			self.form.widgets['file'].show()
		else:
			self.form.widgets['folder'].set_sensitive(True)
			self.form.widgets['folder'].show()
			self.form.widgets['file'].set_sensitive(False)
			self.form.widgets['file'].hide()

	def save_uistate(self):
		self.uistate['output_file'] = self.form['file']
		self.uistate['output_folder'] = self.form['folder']
		self.uistate['index_page'] = self.form['index']
		self.uistate['output'] = self.form['output']


class ExportDoneDialog(MessageDialog):

	def __init__(self, parent, logging_context, output):
		self.logging_context = logging_context
		self.output = output

		n_error = logging_context.handler.n_error
		n_warning = logging_context.handler.n_warning
		if n_error and n_warning:
			text = _('%(n_error)i errors and %(n_warning)i warnings occurred, see log') % {'n_error': n_error, 'n_warning': n_warning}
				# T: label in export dialog
		elif n_error:
			text = _('%i errors occurred, see log') % n_error
				# T: label in export dialog
		elif n_warning:
			text = _('%i warnings occurred, see log') % n_warning
				# T: label in export dialog
		else:
			text = None

		MessageDialog.__init__(self, parent, (_('Export completed'), text))
			# T: label in export dialog

		log_button = Gtk.Button.new_with_mnemonic(_('View _Log'))
			# T: button in export dialog
		log_button.set_sensitive(logging_context.file.exists())
		log_button.connect_object(
			'clicked', self.__class__.on_show_log, self)

		#~ open_button =

		#~ self.add_extra_button(open_button)
		self.add_extra_button(log_button)

	def on_show_log(self):
		LogFileDialog(self, self.logging_context.file).run()

	def on_open_file(self):
		open_file(self.output)


class LogContext(object):
	'''Context to log errors and warnings to a log file'''

	def __init__(self):
		names = ['zim.export', 'zim.templates', 'zim.formats']
		level = logging.INFO

		self.logger = logging.getLogger('zim')
		self.level = level
		self.file = TmpFile(basename='export-log.txt', unique=False, persistent=True)
		self.file.remove() # clean up previous run
		self.file.parent().touch()
		self.handler = LogHandler(self.file.path)
		self.handler.setLevel(self.level)
		self.handler.addFilter(LogFilter(names))
		self.handler.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))

	def __enter__(self):
		#~ self._old_level = self.logger.getEffectiveLevel()
		#~ if self._old_level > self.level:
			#~ self.logger.setLevel(self.level)
		self.logger.addHandler(self.handler)

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.logger.removeHandler(self.handler)
		#~ self.logger.setLevel(self._old_level)
		self.handler.close()
		return False # re-raises error


class LogFilter(logging.Filter):

	def __init__(self, names):
		self.names = names

	def filter(self, record):
		return any(record.name.startswith(n) for n in self.names)


class LogHandler(logging.FileHandler):

	def __init__(self, filename, mode='w', encoding='UTF-8', delay=True):
		logging.FileHandler.__init__(self, filename, mode, encoding, delay)
		self.n_warning = 0
		self.n_error = 0

	def emit(self, record):
		# more detailed logging has lower number, so WARN > INFO > DEBUG
		# log to file unless output is a terminal and logging <= INFO
		if record.levelno >= logging.ERROR:
			self.n_error += 1
		elif record.levelno >= logging.WARNING:
			self.n_warning += 1

		logging.FileHandler.emit(self, record)




# Copyright 2012-2017 Jaap Karssenberg <jaap.karssenberg@gmail.com>


import weakref
import logging
from gi.repository import GObject
import os

from functools import partial


logger = logging.getLogger('zim')

# Constants for signal order
SIGNAL_RUN_FIRST = 1
SIGNAL_BEFORE = SIGNAL_NORMAL = 2
SIGNAL_RUN_LAST = 3
SIGNAL_AFTER = 4



class SignalHandler(object):
	'''Wrapper for a signal handler method that allows blocking the
	handler for incoming signals. To be used as function decorator.

	The method will be replaced by a L{BoundSignalHandler} object that
	supports a C{blocked()} method which returns a context manager
	to temporarily block a callback.

	Intended to be used as::

		class Foo():

			@SignalHandler
			def on_changed(self):
				...

			def update(self):
				with self.on_changed.blocked():
					... # do something that results in a "changed" signal

	When using this handler, make sure to connect the _bound_ method to the signal,
	not the _unbound_ class method.
	'''

	def __init__(self, func):
		self._func = func

	def __get__(self, instance, klass):
		if instance is None:
			# class access
			return self
		else:
			# instance access, return bound version
			name = '_bound_' + self._func.__name__
			if not hasattr(instance, name) \
			or getattr(instance, name) is None:
				bound_obj = BoundSignalHandler(instance, self._func)
				setattr(instance, name, bound_obj)

			return getattr(instance, name)


class BoundSignalHandler(object):

	def __init__(self, instance, func):
		self._instance = instance
		self._func = func
		self._blocked = 0

	def __call__(self, *args, **kwargs):
		if self._blocked == 0:
			return self._func(self._instance, *args, **kwargs)

	def _block(self):
		self._blocked += 1

	def _unblock(self):
		if self._blocked > 0:
			self._blocked -= 1

	def blocked(self):
		'''Returns a context manager that can be used to temporarily
		block a callback.
		'''
		return SignalHandlerBlockContextManager(self)


class SignalHandlerBlockContextManager(object):

	def __init__(self, handler):
		self.handler = handler

	def __enter__(self):
		self.handler._block()

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.handler._unblock()




class ConnectorMixin(object):
	'''Mixin class that has convenience methods for objects that
	want to connect to signals of other objects.
	'''

	# TODO connect variant that wraps with a handler that catches
	#  unwanted arguments -- kind of reverse "partial()"

	def connectto(self, obj, signal, handler=None, order=SIGNAL_NORMAL):
		'''Connect to signals of another object
		E.g.::

			self.connectto(button, 'clicked', self.on_button_clicked)

		@param obj: the object to connect to
		@param signal: the signal name
		@param handler: the callback function, or C{None} to map to
		a method prefixed with "on_".
		@param order: if order is C{SIGNAL_NORMAL} then C{GObject.connect()}
		is used, if order is C{SIGNAL_AFTER} then C{GObject.connect_after()}
		is used.
		@returns: the handler id
		'''
		if handler is None:
			name = "on_" + signal.replace('-', '_')
			handler = getattr(self, name)
			if handler is None:
				raise NotImplementedError('No method "%s"' % name)

		if order == SIGNAL_AFTER:
			i = obj.connect_after(signal, handler)
		else:
			i = obj.connect(signal, handler)

		if not hasattr(self, '_connected_signals'):
			self._connected_signals = {}
			# We might want a dict here that is cleaned up
			# when references disappear, but for now this will do

		key = id(obj)
		if not key in self._connected_signals:
			self._connected_signals[key] = (weakref.ref(obj), [])
		self._connected_signals[key][1].append(i)

		return i

	def connectto_all(self, obj, signals, handler=None, order=SIGNAL_NORMAL):
		'''Convenience method to combine multiple calls to
		L{connectto()}.

		@param obj: the object to connect to
		@param signals: a list of signals. Elements can either be signal
		names or tuples where the sub-elements are the parameters
		for L{connectto()}. For example::

			self.connectto_all(some_object (
				'changed' # defaults to "on_changed()
				('open', on_open_something, None, SIGNAL_AFTER),
			))

		The optional parameters are used as default values when these
		parameters are not specified explicitly per signal.

		@param handler: optional parameter
		@param order: optional parameter
		'''
		default = (None, handler, order)
		for signal in signals:
			if isinstance(signal, str):
				self.connectto(obj, signal, handler, order)
			else:
				arg = signal + default[len(signal):]
					# fill in missing positional arguments
				self.connectto(obj, *arg)

	def disconnect_from(self, obj):
		'''Disc all signals that have been connected with
		L{connectto} and friends to a specific object.
		'''
		key = id(obj)
		if hasattr(self, '_connected_signals') \
		and key in self._connected_signals:
			self._disconnect_from(key)

	def disconnect_all(self):
		'''Disconnect all signals that have been connected with
		L{connectto} and friends. Typically called when you want to
		destroy this object.
		'''
		if hasattr(self, '_connected_signals'):
			for key in list(self._connected_signals.keys()):
				try:
					self._disconnect_from(key)
				except:
					logger.exception('Exception in disconnect_all()')

	def _disconnect_from(self, key):
		ref, signals = self._connected_signals[key]
		obj = ref()
		if obj is not None:
			for i in signals:
				obj.disconnect(i)
		del self._connected_signals[key]


class SignalEmitterMeta(type):

	def __init__(cls, name, bases, dct):
		# Init class

		assert hasattr(cls, '__signals__')
		assert not hasattr(cls, '__gsignals__'), 'Did you mean to define "__signals__" instead of "__gsignals__"?'

		#  1/ setup inheritance for signals

		if name != 'SignalEmitter':
			for base in bases:
				if issubclass(base, SignalEmitter):
					for key, value in list(base.__signals__.items()):
						cls.__signals__.setdefault(key, value)

		#  2/ set list of closures to be initialized per instance

		cls._signal_closures = []
		for signal in cls.__signals__:
			name = 'do_' + signal.replace('-', '_')
			if hasattr(cls, name):
				order = cls.__signals__[signal][0]
				if not order in (SIGNAL_RUN_FIRST, SIGNAL_RUN_LAST):
					order = SIGNAL_RUN_LAST # for backward compatibility, fallback to this default
				closure = getattr(cls, name) # unbound version!
				cls._signal_closures.append((signal, order, closure))

		super(SignalEmitterMeta, cls).__init__(name, bases, dct)


def init_signals_for_new_object(obj):
	obj._signal_handlers = {}
	obj._signal_blocks = {}
	obj._signal_count = 0 # ensure signals execute in order of connecting

	for signal, order, closure in obj._signal_closures:
		obj._signal_handlers[signal] = [(order, 0, closure)]


class SignalEmitter(object, metaclass=SignalEmitterMeta):
	'''Replacement for C{GObject} to make objects emit signals.
	API should be (mostly) compatible with API offered by GObject.

	Supported signals need to be defined in the dict C{__signals__}. For
	each signal a 3-tuple is provided where the first argument is either
	C{SIGNAL_RUN_FIRST} or C{SIGNAL_RUN_LAST}, the second is the return
	argument (or C{None} for most signals) and the third is the argument
	spec for the signal.

	When a signal is emitted all handlers are called one by one and exceptions
	in a handler are intercepted. If the class defines a method
	C{do_signalname()} (where "signalname" is the name of the signal with "-"
	replaced by "_") this is considered a default handler that is automatically
	connected. The C{SIGNAL_RUN_FIRST} and C{SIGNAL_RUN_LAST} flags in the
	signal spec control where in the  emission sequence this default handler
	is called: it either runs as the first or as the last handler.
	Typically the default handler is used to implement the change signalled by
	the signal. So if your signal name suggests something happened already
	(e.g. "changed"), probably you want to use C{SIGNAL_RUN_FIRST}. However if
	the action is phrased as still happening (e.g. "change") and the handlers
	can influence this, you probably want C{SIGNAL_RUN_LAST}.

	Signals connected with L{connect_after()} always run after all the normal
	handlers (which are connected with L{connect()}) and after the default handler
	(even if the default handler is setup with C{SIGNAL_RUN_LAST}).

	If a signal defines a return type, emission should use either
	L{emit_return_first()} or L{emit_return_iter()} to handle return values.

	(Also see the Glib documentation to understand the full system of which
	we implement a sub-set here.)
	'''

	# define signals we want to use - (closure type, return type and arg types)
	# E.g. {signal: (SIGNAL_RUN_LAST, None, (object, object))}
	__signals__ = {} #: signals supported by this class

	def __new__(cls, *arg, **kwarg):
		# New instance: init attributes for signal handling
		obj = super(SignalEmitter, cls).__new__(cls)
		init_signals_for_new_object(obj)
		return obj

	def connect(self, signal, handler):
		'''Register a handler for a specific object.

		Note that connecting makes a hard reference to the connected
		object. So connecting an bound method will prevent the
		object the method belongs to to be destroyed until the
		signal is disconnected.

		@param signal: the signal name
		@param handler: callback to be called upon the signal,
		first object to the callback will be the emitting object,
		other params are signal specific.
		@returns: an id for the registered handler
		'''
		return self._connect(SIGNAL_BEFORE, signal, handler)

	def connect_after(self, signal, handler):
		'''Like L{connect()} but handler will be called after default handler'''
		return self._connect(SIGNAL_AFTER, signal, handler)

	def _connect(self, category, signal, callback):
		assert signal in self.__signals__, 'No such signal: %s::%s' % (self.__class__.__name__, signal)

		if not signal in self._signal_handlers:
			self._signal_handlers[signal] = []
			self._setup_signal(signal)

		# The "handler" is a tuple with an unique object id, which is
		# used as the handler id. It starts with a category and a counter
		# to ensure sorting is stable, both for categories and for the
		# order of connecting
		handler = (category, self._signal_count, callback)
		self._signal_count += 1
		self._signal_handlers[signal].append(handler)
		self._signal_handlers[signal].sort()
		return id(handler)

	def _setup_signal(self, signal):
		pass

	def disconnect(self, handlerid):
		for signal, handlers in list(self._signal_handlers.items()):
			# unique id, so when we find it, stop searching
			ids = list(map(id, handlers))
			try:
				i = ids.index(handlerid)
			except ValueError:
				continue
			else:
				handlers.pop(i)
				if not handlers:
					self._signal_handlers.pop(signal)
					self._teardown_signal(signal)
				return True
		else:
			return False

	def _teardown_signal(self, signal):
		pass

	def emit(self, signal, *args):
		# NOTE: do *not* refactor this method as a wrapper around
		# emit_return_iter() or similar. It is called often enough to justify
		# being optimized at the cost of some redundant code.
		assert signal in self.__signals__, 'No such signal: %s::%s' % (self.__class__.__name__, signal)

		if not len(args) == len(self.__signals__[signal][2]):
			logger.warning('Signal args do not match spec for %s::%s', self.__class__.__name__, signal)

		assert self.__signals__[signal][1] is None, 'This signal expects return values, use either emit_return_first() or emit_return_iter()'

		if self._signal_blocks.get(signal):
			return # ignore emit

		for c, i, handler in self._signal_handlers.get(signal, []):
			try:
				r = handler(self, *args)
			except:
				logger.exception('Exception in signal handler for %s on %s', signal, self)

	def emit_return_first(self, signal, *args):
		'''Emits a signal and stops emission on the first handler that returns
		a not-None value.
		'''
		for r in self.emit_return_iter(signal, *args):
			if r is not None:
				return r

	def emit_return_iter(self, signal, *args):
		'''Returns an generator that calls one handler on each iteration and
		yields the return values. This allows aggregating return values.
		'''
		assert signal in self.__signals__, 'No such signal: %s::%s' % (self.__class__.__name__, signal)

		if not len(args) == len(self.__signals__[signal][2]):
			logger.warning('Signal args do not match spec for %s::%s', self.__class__.__name__, signal)

		if not self._signal_blocks.get(signal):
			for c, i, handler in self._signal_handlers.get(signal, []):
				try:
					yield handler(self, *args)
				except GeneratorExit:
					raise
				except:
					logger.exception('Exception in signal handler for %s on %s', signal, self)

	def block_signals(self, *signals):
		'''Returns a context manager for blocking one or more signals'''
		return BlockSignalsContextManager(self, signals)


class GSignalEmitterMixin(object):
	'''Implements a subset of L{SignalEmitter} to extend C{GObject.GObject}
	classes with methods to use signals as callbacks.
	'''

	def __init__(self):
		self._signal_hooks = [
			s for s, k in list(self.__signals__.items())
				if k[1] is not None
		]
		clsname = self.__class__.__name__ + 'SignalEmitter'
		signals = dict(
			(k, (SIGNAL_RUN_FIRST,) + v[1:])
				for k, v in list(self.__signals__.items()) if k in self._signal_hooks
		)
		innercls = type(clsname, (SignalEmitter,), {'__signals__': signals})
		self._signals_inner = innercls()

		for signal in self._signal_hooks:
			methodname = 'do_' + signal.replace('-', '_')
			if hasattr(self, methodname):
				self._signals_inner.connect_after(
					signal,
					lambda o, *a: getattr(self, methodname)(*a)
				)

	def connect(self, signal, *args):
		if signal in self._signal_hooks:
			return self._signals_inner.connect(signal, *args)
		else:
			return GObject.GObject.connect(self, signal, *args)

	def connect_after(self, signal, *args):
		if signal in self._signal_hooks:
			return self._signals_inner.connect_after(signal, *args)
		else:
			return GObject.GObject.connect_after(self, signal, *args)

	def disconnect(self, id):
		self._signals_inner.disconnect(id) \
			or GObject.GObject.disconnect(self, id)

	def emit_return_first(self, signal, *args):
		return self._signals_inner.emit_return_first(signal, *args)

	def emit_return_iter(self, signal, *args):
		return self._signals_inner.emit_return_iter(signal, *args)


class BlockSignalsContextManager(object):

	def __init__(self, obj, signals):
		self.obj = obj
		self.signals = signals

	def __enter__(self):
		for signal in self.signals:
			self.obj._signal_blocks.setdefault(signal, 0)
			self.obj._signal_blocks[signal] += 1

	def __exit__(self, exc_type, exc_val, exc_tb):
		for signal in self.signals:
			if signal in self.obj._signal_blocks \
			and self.obj._signal_blocks[signal] > 0:
				self.obj._signal_blocks[signal] -= 1



class DelayedCallback(object):
	'''Wrapper for callbacks that need to be delayed after a signal

	This class allows you to add a callback to a signal, but only have
	it called after a certain timeout. If the signal is emitted
	again during this time the callback will be canceled and the
	timeout starts again. (So the callback is not called for each repeat
	of the signal.) This can be used e.g. in case want to update some
	other widget after the user changes a text entry widget, but this
	can be done once the user pauses, while calling the callback for
	every key stroke would make the application non-responsive.

	Objects of this class wrap the actual callback function and can be
	called as a normal function.

	Note that when a repeated callback is canceled, only the arguments
	of the last call are passed on.

	@todo: allow an option to check arguments and pass on all unique
	combinations ?

	@todo: add support for async callbacks, in this case block
	the callback until the async process is finished
	'''

	__slots__ = ('timeout', 'cb_func', 'timer_id')

	def __init__(self, timeout, cb_func):
		'''Constructor

		@param timeout: timeout in milliseconds (e.g. 500)
		@param cb_func: the callback to call
		'''
		self.cb_func = cb_func
		self.timeout = timeout
		self.timer_id = None

	def __call__(self, *arg, **kwarg):
		if self.timer_id:
			GObject.source_remove(self.timer_id)
			self.timer_id = None

		def callback():
			self.timer_id = None
			self.cb_func(*arg, **kwarg)
			return False # destroy timeout

		self.timer_id = GObject.timeout_add(self.timeout, callback)

	def __del__(self):
		if self.timer_id:
			GObject.source_remove(self.timer_id)

	def cancel(self):
		'''Cancel the scheduled callback'''
		self.__del__()


def callback(func, *arg, **kwarg):
	'''Returns a wrapper functions that call func and.
	Intended as wrapper for callbacks connected to (gtk) signals. The
	wrapper ignores any arguments given.
	'''
	def cb(*a):
		func(*arg, **kwarg)
	return cb




# Copyright 2009-2012 Jaap Karssenberg <jaap.karssenberg@gmail.com>
# Copyright 2012 Damien Accorsi <damien.accorsi@free.fr>



import os
import logging

from zim.plugins.versioncontrol import VCSApplicationBase
from zim.applications import Application, ApplicationError


logger = logging.getLogger('zim.vcs.bzr')


class BZRApplicationBackend(VCSApplicationBase):

	@classmethod
	def build_bin_application_instance(cls):
		return Application(('bzr',))

	def build_revision_arguments(self, versions):
		"""Build a list including required string/int for running an VCS command
		# Accepts: None, int, string, (int,), (int, int)
		# Always returns a list
		# versions content:
		  - None: return an empty list
		  - int ou string: return ['-r', int]
		  - tuple or list: return ['-r', '%i..%i']

		It's all based on the fact that defining revision with current VCS is:
		-r revision
		-r rev1..rev2
		"""
		if isinstance(versions, (tuple, list)):
			assert 1 <= len(versions) <= 2
			if len(versions) == 2:
				versions = sorted(map(int, versions))
				return ['-r', '%i..%i' % tuple(versions)]
			else:
				versions = versions[0]

		if not versions is None:
			version = int(versions)
			return ['-r', '%i' % version]
		else:
			return []

	########
	#
	# NOW ARE ALL REVISION CONTROL SYSTEM SHORTCUTS

	def add(self, path=None):
		"""
		Runs: bzr add {{PATH}}
		"""
		if path is None:
			return self.run(['add', self.notebook_dir])
		else:
			return self.run(['add', path])


	def annotate(self, file, version=None):
		"""FIXME Document
		return
		1 | line1
		2 | line2
		...
		"""
		revision_args = self.build_revision_arguments(version)
		return self.pipe(['annotate', file] + revision_args)

	def cat(self, path, version):
		"""
		Runs: bzr cat {{PATH}} {{REV_ARGS}}
		"""
		revision_args = self.build_revision_arguments(version)
		return self.pipe(['cat', path] + revision_args)

	def commit(self, path, msg):
		"""
		Runs: bzr commit -m {{MSG}} {{PATH}}
		"""
		params = ['commit']
		if msg != '' and msg is not None:
			params.append('-m')
			params.append(msg)
		if path != '' and path is not None:
			params.append(path)
		return self.run(params)

	def diff(self, versions=None, file=None):
		"""
		Runs:
			bzr diff {{REVISION_ARGS}}
		or
			bzr diff {{REVISION_ARGS}} {{PATH}}
		"""
		revision_args = self.build_revision_arguments(versions)
		if file is None:
			return self.pipe(['diff'] + revision_args)
			# Using --git option allow to show the renaming of files
		else:
			return self.pipe(['diff', file] + revision_args)

	def ignore(self, file_to_ignore_regexp):
		"""
		Build a .bzrignore file including the file_to_ignore_content
		"""
		return self.run(['ignore', file_to_ignore_regexp])


	def init_repo(self):
		self.init()
		if not self.test_whoami():
			self.whoami('zim') # set a dummy user "zim"
		self.ignore('**/.zim/')
		self.add('.')

	def repo_exists(self):
		return self.root.folder('.bzr').exists()

	def init(self):
		"""
		Runs: bzr init
		"""
		return self.run(['init'])

	def is_modified(self):
		"""Returns true if the repo is not up-to-date, or False
		@returns: True if the repo is not up-to-date, or False
		"""
		# If status return an empty answer, this means the local repo is up-to-date
		return ''.join(self.status()).strip() != ''

	def log(self, path=None):
		"""
		Runs: bzr log --forward {{PATH}}
		the "--forward" option allows to reverse order
		"""
		if path:
			return self.pipe(['log', '--forward', path])
		else:
			return self.pipe(['log', '--forward'])

	def log_to_revision_list(self, log_op_output):
		versions = []
		(rev, date, user, msg) = (None, None, None, None)
		seenmsg = False
		# seenmsg allow to get the complete commit message which is presented like this:
		#
		# [...]
		# description:
		# here is the
		# commit message
		# the end of it may be detected
		# because of the apparition of a line
		# starting by "changeset:"
		#
		# FIXME: there is a bug which will stop parsing if a blank line is included
		# in the commit message
		for line in log_op_output:
			if line.startswith('----'):
				if not rev is None:
					versions.append((rev, date, user, msg))
				(rev, date, user, msg) = (None, None, None, None)
			elif line.startswith('revno: '):
				value = line[7:].strip()
				if ' ' in value:
					# e.g. "revno: 48 [merge]\n"
					i = value.index(' ')
					value = value[:i]
				rev = value
			elif line.startswith('committer: '):
				user = line[11:].strip()
			elif line.startswith('timestamp: '):
				date = line[11:].strip()
			elif line.startswith('message:'):
				seenmsg = True
				msg = ''
			elif seenmsg and line.startswith('  '):
				msg += line[2:]

		if not rev is None:
			versions.append((rev, date, user, msg))

		return versions


	def move(self, oldpath, newpath):
		"""
		Runs: bzr mv --after {{OLDPATH}} {{NEWPATH}}
		"""
		self.run(['add', '--no-recurse', newpath.dirname])
		return self.run(['mv', oldpath, newpath])


	def remove(self, path):
		"""
		Runs: bzr rm {{PATH}}
		"""
		return self.run(['rm', path])

	def revert(self, path=None, version=None):
		"""
		Runs:
			bzr revert {{PATH}} {{REV_ARGS}}
		or
			bzr revert {{REV_ARGS}}
		"""
		revision_params = self.build_revision_arguments(version)
		if path:
			return self.run(['revert', path] + revision_params)
		else:
			return self.run(['revert'] + revision_params)

	def status(self):
		"""
		Runs: bzr status
		"""
		return self.pipe(['status'])

	def whoami(self, user):
		"""
		Runs: bzr whoami zim
		"""
		return self.pipe(['whoami', user])

	def test_whoami(self):
		"""return True if the user is is setup or non-zero
		"""
		try:
			self.run(['whoami'])
		except ApplicationError as e:
			return False
		else:
			return True



# -*- coding: utf-8 -*-

# Copyright 2020 Jaap Karssenberg <jaap.karssenberg@gmail.com>

from zim.gui.pageview import PageView, PageViewExtension, PageViewExtensionBase
from zim.plugins import extendable


class NotebookViewExtension(PageViewExtensionBase):
	'''Base class for extensions of the L{NotebookView},
	see L{PageViewExtensionBase} for API documentation.
	'''
	pass


@extendable(NotebookViewExtension, PageViewExtension, register_after_init=False)
class NotebookView(PageView):
	'''Sub-class of the L{PageView} class that is used when the view is
	navigatable. Plugins can choose to extend this class if they don't want
	to load extensions for a fixed page view.
	'''
	pass
